import datetime
import os
import random
import re
import sys
import typing
from copy import copy

from pymead.core.param_graph import ParamGraph
from pymead.core import UNITS
from pymead.core.airfoil import Airfoil
from pymead.core.bezier import Bezier
from pymead.core.constraints import *
from pymead.core.gcs import GCS
from pymead.core.mea import MEA
from pymead.core.pymead_obj import DualRep, PymeadObj
from pymead.core.line import LineSegment, PolyLine
from pymead.core.param import Param, LengthParam, AngleParam, DesVar, LengthDesVar, AngleDesVar
from pymead.core.point import Point, PointSequence
from pymead.core.transformation import Transformation3D
from pymead.plugins.IGES.curves import BezierIGES
from pymead.plugins.IGES.iges_generator import IGESGenerator
from pymead.version import __version__


class GeometryCollection(DualRep):
    def __init__(self, gui_obj=None):
        """
        The geometry collection is the primary class in pymead for housing all the available fundamental geometry
        types. Geometry, parameters, and constraints can be added using the nomenclature ``add_<object-name>()``.
        For example, points can be added using ``geo_col.add_point(x=<x-value>, y=<y-value>)``.
        """
        self._container = {
            "desvar": {},
            "params": {},
            "points": {},
            "lines": {},
            "polylines": {},
            "bezier": {},
            "airfoils": {},
            "mea": {},
            "geocon": {},
        }
        self.gcs = GCS()
        self.param_graph = ParamGraph()
        self.gcs.geo_col = self
        self.gui_obj = gui_obj
        self.canvas = None
        self.tree = None
        self.selected_objects = {k: [] for k in self._container.keys()}
        self.selected_airfoils = []
        self.single_step = 0.01

    def container(self):
        """
        Retrieves the geometry container. Note that there is no "setter" method because geometry items should
        be added only be their respective methods. For example, a ``Param`` should be added using ``.add_param()``.
        This allows for name validation, which ensures that every geometry item in the container has a unique name.

        Returns
        =======
        dict
            Dictionary of geometry items
        """
        return self._container

    def clear_container(self):
        """
        Clears all the entities in the geometry container.
        """
        self.container().clear()

    def get_name_list(self, sub_container: str):
        """
        Gets a list of all the parameter or geometry names in a specified sub-container.

        Parameters
        ==========
        sub_container: str
            Sub-container in the geometry collection. For example, ``"params"``, ``"lines"``, ``"points"``, etc.

        Returns
        =======
        typing.List[str]
            List of names found in the sub-container
        """
        return [k for k in self.container()[sub_container].keys()]

    @staticmethod
    def unique_namer(specified_name: str, name_list: typing.List[str]):
        """
        This static method creates unique names for parameters or geometry by incrementing the appended index, which
        is attached with a hyphen. A dot separator is used to distinguish between levels within the geometry
        collection hierarchy. For example, the :math:`x`-location of the second point added is given by
        ``"Point-2.x"``. Note that, for parameters and design variables, the first parameter added with a given name
        will not have an index by default. For example, if a ``Param`` with ``name=="my_param"`` is added three times,
        the resulting names, in order, will be ``"my_param"``, ``"my_param-2"``, and ``"my_param-3"``.

        If a parameter is removed, and another
        parameter with the same name added, the index assigned will be the maximum index plus one. In the previous
        example, if ``"my_param-2"`` is removed from the geometry collection and the addition of ``"my_param"``
        requested, the actual name assigned will be ``"my_param-4"``. This is to prevent confusion between removed and
        added parameters, as well as to give an indication of the order in which the parameter was added.

        Parameters
        ==========
        specified_name: str
            Input name to be tested or modified for uniqueness. Does not need to have an index.

        name_list: typing.List[str]
            List of names extracted from a particular sub-container in the geometry collection, used to check the
            ``specified_name`` for uniqueness.

        Returns
        =======
        str
            The (possibly modified) ``specified_name``
        """
        specified_name = specified_name.split("-")[0]
        max_index = 0
        for name in name_list:
            dot_split = name.split(".")
            dash_split = dot_split[-1].split("-")
            if len(dot_split) == 1:  # If there were no dots,
                prefix = dash_split[0]
            else:
                prefix = ".".join(dot_split)[:-1] + dash_split[0]
            if specified_name != prefix:  # If the specified name does not match the name
                # being analyzed, continue to next name
                continue
            if len(dash_split) < 2 and max_index == 0:  # If the string after the last dot had no hyphen
                # and max_index is still 0, set the max_index to 1.
                max_index = 1
            elif len(dash_split) >= 2:  # If the string after the last dot had a hyphen and the index is greater than
                # the max_index, set the max_index to this index.
                idx = int(dash_split[-1])
                if idx > max_index:
                    max_index = idx

        if max_index == 0:
            return f"{specified_name}-1"
        else:
            return f"{specified_name}-{max_index + 1}"

    def add_to_subcontainer(self, pymead_obj: PymeadObj, assign_unique_name: bool = True):
        """
        Adds an object to a sub-container within the geometry collection's ``container()``. Also performs the task
        of assigning a unique name to the object before insertion into the sub-container, if necessary.
        
        Parameters
        ==========
        pymead_obj: PymeadObj
            Object to add to the sub-container
            
        assign_unique_name: bool
            Whether to assign the object a unique name before insertion into the sub-container. Default: ``True``.
        """
        # Set the object's name to a unique name if necessary
        if assign_unique_name:
            name_list = self.get_name_list(sub_container=pymead_obj.sub_container)
            unique_name = self.unique_namer(pymead_obj.name(), name_list)
            if isinstance(pymead_obj, Param) and unique_name.split("-")[1] == "1":
                pass
            else:
                pymead_obj.set_name(unique_name)

            if isinstance(pymead_obj, GeoCon):
                if pymead_obj.param() is not None and pymead_obj.param().name() == "unnamed":
                    pymead_obj.param().set_name(f"{pymead_obj.name()}.par")

        # Add the object to the geometry collection sub-container
        self.container()[pymead_obj.sub_container][pymead_obj.name()] = pymead_obj

    def remove_from_subcontainer(self, pymead_obj: PymeadObj):
        """
        Removes an object from the specified sub-container.

        Parameters
        ==========
        pymead_obj: PymeadObj
            Object to remove.
        """
        if pymead_obj.name() not in self.container()[pymead_obj.sub_container]:
            return
        self.container()[pymead_obj.sub_container].pop(pymead_obj.name())

    def add_param(self, value: float, name: str or None = None, lower: float or None = None,
                  upper: float or None = None, unit_type: str or None = None, assign_unique_name: bool = True,
                  point: Point = None, root: Point = None, rotation_handle: Point = None, enabled: bool = True,
                  equation_str: str = None):
        """
        Adds a parameter to the geometry collection sub-container ``"params"``, and modifies the name to make it
        unique if necessary.

        Parameters
        ==========
        value: float
            Parameter value

        name: str or None
            Parameter name

        lower: float or None
            Lower bound. If ``None``, no bound will be set. Default: ``None``.

        upper: float or None
            Upper bound. If ``None``, no bound will be set. Default: ``None``.

        unit_type: str or None:
            The unit type of design variable to create. Default: ``None``.

        Returns
        =======
        Param
            The generated parameter
        """
        kwargs = dict(value=value, name=name, lower=lower, upper=upper, setting_from_geo_col=True, point=point,
                      root=root, rotation_handle=rotation_handle, enabled=enabled, equation_str=equation_str)
        if unit_type is None:
            param = Param(**kwargs)
        elif unit_type == "length":
            param = LengthParam(**kwargs)
        elif unit_type == "angle":
            param = AngleParam(**kwargs)
        else:
            raise ValueError(f"unit_type must be None, 'length', or 'angle'. Found type: {type(unit_type)}")

        return self.add_pymead_obj_by_ref(param, assign_unique_name=assign_unique_name)

    def select_object(self, pymead_obj: PymeadObj):
        if self.tree is not None:
            self.tree.setItemStyle(pymead_obj.tree_item, "default")
            pymead_obj.tree_item.hoverable = False
            pymead_obj.tree_item.setSelected(True)

        if self.canvas is not None:
            if isinstance(pymead_obj, Point):
                self.canvas.setItemStyle(pymead_obj.canvas_item, "selected")
                pymead_obj.canvas_item.hoverable = False
            elif isinstance(pymead_obj, Airfoil):
                for curve in pymead_obj.curves:
                    self.canvas.setItemStyle(curve.canvas_item, "hovered")
                    curve.canvas_item.hoverable = False

        # if isinstance(pymead_obj, Point):
        #     if pymead_obj not in self.selected_objects:
        #         self.selected_objects.append(pymead_obj)
        # elif isinstance(pymead_obj, Airfoil):
        #     if pymead_obj not in self.selected_airfoils:
        #         self.selected_airfoils.append(pymead_obj)

        if pymead_obj not in self.selected_objects[pymead_obj.sub_container]:
            self.selected_objects[pymead_obj.sub_container].append(pymead_obj)

        # print(f"Selecting object {pymead_obj}. {self.selected_objects = }")

    def deselect_object(self, pymead_obj: PymeadObj):
        if self.tree is not None:
            if pymead_obj.tree_item is not None:
                pymead_obj.tree_item.hoverable = True
                pymead_obj.tree_item.setSelected(False)

        if self.canvas is not None:
            if isinstance(pymead_obj, Point):
                pymead_obj.canvas_item.hoverable = True
                self.canvas.setItemStyle(pymead_obj.canvas_item, "default")
            elif isinstance(pymead_obj, LineSegment):
                pymead_obj.canvas_item.hoverable = True
                self.canvas.setItemStyle(pymead_obj.canvas_item, "default")
            elif isinstance(pymead_obj, Bezier):
                pymead_obj.canvas_item.hoverable = True
                self.canvas.setItemStyle(pymead_obj.canvas_item, "default")
            elif isinstance(pymead_obj, Airfoil):
                for curve in pymead_obj.curves:
                    curve.canvas_item.hoverable = True
                    self.canvas.setItemStyle(curve.canvas_item, "default")
            elif isinstance(pymead_obj, GeoCon):
                pymead_obj.canvas_item.hoverable = True
                pymead_obj.canvas_item.setStyle(theme=self.gui_obj.themes[self.gui_obj.current_theme])

        # if isinstance(pymead_obj, Point):
        #     if pymead_obj in self.selected_objects:
        #         self.selected_objects.remove(pymead_obj)
        # elif isinstance(pymead_obj, Airfoil):
        #     if pymead_obj in self.selected_airfoils:
        #         self.selected_airfoils.remove(pymead_obj)

        if pymead_obj in self.selected_objects[pymead_obj.sub_container]:
            self.selected_objects[pymead_obj.sub_container].remove(pymead_obj)

    def clear_selected_objects(self):
        # for point in self.selected_objects[::-1]:
        #     self.deselect_object(point)
        # for airfoil in self.selected_airfoils[::-1]:
        #     self.deselect_object(airfoil)

        for d in self.selected_objects.values():
            for obj in d[::-1]:
                self.deselect_object(obj)

    def remove_selected_objects(self):
        # Remove only the points first for speed (points are the core object in pymead, so deleting a point deletes
        # all associated pymead objects)
        for pt in self.selected_objects["points"]:
            self.remove_pymead_obj(pt)

        # Remove all the other selected objects
        remaining_subcontainers = [k for k in self.container().keys() if k != "points"]
        for sub_container in remaining_subcontainers:
            for obj in self.selected_objects[sub_container]:
                self.remove_pymead_obj(obj)

        # Clear the selected objects
        self.clear_selected_objects()

    def hover_enter_obj(self, pymead_obj: PymeadObj):
        if self.tree is not None:
            self.tree.setItemStyle(pymead_obj.tree_item, "hovered")

        if self.canvas is not None:
            self.canvas.setItemStyle(pymead_obj.canvas_item, "hovered")

    def hover_leave_obj(self, pymead_obj: PymeadObj):
        if self.tree is not None:
            self.tree.setItemStyle(pymead_obj.tree_item, "default")

        if self.canvas is not None:
            self.canvas.setItemStyle(pymead_obj.canvas_item, "default")

    def add_pymead_obj_by_ref(self, pymead_obj: PymeadObj, assign_unique_name: bool = True):
        """
        This method adds a pymead object by passing it directly to the geometry collection. If the
        object is already associated with a geometry collection, a ``ValueError`` is raised.

        Parameters
        ==========
        pymead_obj: PymeadObj
            The pymead object to add to the collection

        Returns
        =======
        PymeadObj
            The modified pymead object
        """
        if pymead_obj.geo_col is not None:
            if isinstance(pymead_obj, Param) and pymead_obj.point is not None:
                pass
            else:
                raise ValueError("Can only add a pymead object by reference if it has not yet been added to a "
                                 "geometry collection")

        pymead_obj.geo_col = self

        self.add_to_subcontainer(pymead_obj, assign_unique_name=assign_unique_name)

        if self.gui_obj is not None:
            pymead_obj.gui_obj = self.gui_obj

        if self.tree is not None:
            pymead_obj.tree = self.tree
            self.tree.addPymeadTreeItem(pymead_obj=pymead_obj)

        if self.canvas is not None:
            pymead_obj.canvas = self.canvas
            self.canvas.addPymeadCanvasItem(pymead_obj=pymead_obj)

        if isinstance(pymead_obj, Point):
            self.gcs.add_point(pymead_obj)
            for param in [pymead_obj.x(), pymead_obj.y()]:
                param.param_graph = self.param_graph
                if param not in param.param_graph.param_list:
                    param.param_graph.param_list.append(param)
                param.update_equation(param.equation_str)

        if isinstance(pymead_obj, Param):
            pymead_obj.param_graph = self.param_graph
            if pymead_obj not in pymead_obj.param_graph.param_list:
                pymead_obj.param_graph.param_list.append(pymead_obj)
            pymead_obj.update_equation(pymead_obj.equation_str)
            pymead_obj.set_enabled(pymead_obj.enabled())

        return pymead_obj

    def remove_pymead_obj(self, pymead_obj: PymeadObj, promotion_demotion: bool = False,
                          constraint_removal: bool = False, equating_constraints: bool = False):
        """
        Removes a pymead object from the geometry collection.

        Parameters
        ==========
        pymead_obj: PymeadObj
            Pymead object to remove

        promotion_demotion: bool
            When this flag is set to ``True``, the ``ValueError`` normally raised when directly deleting a ``Param``
            associated with a ``GeoCon`` is ignored. Default: ``False``

        constraint_removal: bool
            When this flag is set to ``True``, the ``ValueError`` normally raise when directly deleting a ``Param``
            associated with a constraint cluster rotation is ignored. Default: ``False``

        equating_constraints: bool
            When this flag is set to ``True`` and the ``pymead_obj`` is a ``Param``, the associated constraints are
            not deleted
        """
        # Type-specific actions
        if isinstance(pymead_obj, Param):

            if pymead_obj.rotation_handle and not constraint_removal and not promotion_demotion:
                error_message = f"This parameter can only be removed by deleting its associated constraint cluster"
                if self.gui_obj is None:
                    raise ValueError(error_message)
                else:
                    self.gui_obj.disp_message_box(error_message, message_mode="error")
                    return

            if not promotion_demotion and not equating_constraints:  # Do not remove the constraints if this is a
                # promotion/demotion action or an equating constraints action
                for geo_con in pymead_obj.geo_cons:
                    self.remove_pymead_obj(geo_con)

            if pymead_obj in self.param_graph.param_list:
                self.param_graph.param_list.remove(pymead_obj)
            if pymead_obj in self.param_graph.nodes:
                self.param_graph.remove_node(pymead_obj)

        elif isinstance(pymead_obj, Bezier) or isinstance(pymead_obj, LineSegment) or isinstance(pymead_obj, PolyLine):
            # Remove all the references to this curve in each of the curve's points
            for pt in pymead_obj.point_sequence().points():
                if pymead_obj in pt.curves:
                    pt.curves.remove(pymead_obj)

            if pymead_obj.airfoil is not None:
                for curve in pymead_obj.airfoil.curves:
                    if pymead_obj is not curve:
                        curve.airfoil = None
                self.remove_pymead_obj(pymead_obj.airfoil)

        elif isinstance(pymead_obj, Point):

            for param in [pymead_obj.x(), pymead_obj.y()]:
                param.param_graph.param_list.remove(param)
                if param in param.param_graph.nodes:
                    param.param_graph.remove_node(param)

            for geo_con in pymead_obj.geo_cons:
                self.remove_pymead_obj(geo_con)

            # Loop through the curves associated with this point to see which ones need to be deleted if one point
            # is removed from their point sequence
            curves_to_delete = []
            for curve in pymead_obj.curves:
                if curve.point_removal_deletes_curve():
                    curves_to_delete.append(curve)

            # Remove the curves that need to be removed due to insufficient points in the point sequence
            for curve in curves_to_delete:
                try:
                    self.remove_pymead_obj(curve)
                except KeyError:
                    pass

            # Update any remaining curves
            for curve in pymead_obj.curves:
                if pymead_obj in curve.point_sequence().points():
                    curve.remove_point(point=pymead_obj)
                    curve.update()

            for geo_con in pymead_obj.geo_cons[::-1]:
                self.remove_pymead_obj(geo_con)

            self.gcs.remove_point(pymead_obj)

        elif isinstance(pymead_obj, GeoCon):
            # First, remove the parameter associated with the constraint if necessary (i.e., if that parameter is not
            # tied to any other constraints
            if pymead_obj.param() is not None:
                pymead_obj.param().geo_cons.remove(pymead_obj)
                if len(pymead_obj.param().geo_cons) == 0:
                    self.remove_pymead_obj(pymead_obj.param())

            # if (isinstance(pymead_obj, DistanceConstraint) or isinstance(pymead_obj, RelAngle3Constraint) or
            #     isinstance(pymead_obj, Perp3Constraint) or isinstance(pymead_obj, AntiParallel3Constraint)):
            #     if pymead_obj.p2.rotation_handle and pymead_obj.p2.rotation_param is not None:
            #         self.remove_pymead_obj(pymead_obj.p2.rotation_param, constraint_removal=True)

            # Remove the constraint from the ConstraintGraph
            self.gcs.remove_constraint(pymead_obj)

        elif isinstance(pymead_obj, Airfoil):
            for curve in pymead_obj.curves:
                curve.airfoil = None
            mea_to_delete = []
            for mea in self.container()["mea"].values():
                if pymead_obj not in mea.airfoils:
                    continue
                mea_to_delete.append(mea)
            for mea in mea_to_delete:
                self.remove_pymead_obj(mea)

        # Remove the item from the geometry collection subcontainer
        self.remove_from_subcontainer(pymead_obj)

        # Remove the tree item if it exists
        if self.tree is not None:
            self.tree.removePymeadTreeItem(pymead_obj)

        # Remove the canvas item if it exists
        if self.canvas is not None:
            if hasattr(pymead_obj.canvas_item, "canvas_items"):  # This is the case for GeoCons
                for canvas_item in pymead_obj.canvas_item.canvas_items:
                    self.canvas.removeItem(canvas_item)
            else:
                self.canvas.removeItem(pymead_obj.canvas_item)

        if isinstance(pymead_obj, Airfoil):
            self.gui_obj.permanent_widget.updateAirfoils()

    def add_point(self, x: float, y: float, name: str or None = None, assign_unique_name: bool = True):
        """
        Adds a point by value to the geometry collection

        Parameters
        ==========
        x: float
            :math:`x`-location of the point

        y: float
            :math:`y`-location of the point

        name: str
            Optional name for the point

        Returns
        =======
        Point
            Object reference
        """
        point = Point(x=x, y=y, name=name)
        point.x().geo_col = self
        point.y().geo_col = self
        self.add_pymead_obj_by_ref(point, assign_unique_name=assign_unique_name)

        return point

    def add_bezier(self, point_sequence: PointSequence, name: str or None = None,
                   t_start: float = None, t_end: float = None, assign_unique_name: bool = True):
        bezier = Bezier(point_sequence=point_sequence, name=name, t_start=t_start, t_end=t_end)

        return self.add_pymead_obj_by_ref(bezier, assign_unique_name=assign_unique_name)

    def add_line(self, point_sequence: PointSequence, name: str or None = None, assign_unique_name: bool = True):
        line = LineSegment(point_sequence=point_sequence, name=name)

        return self.add_pymead_obj_by_ref(line, assign_unique_name=assign_unique_name)

    def add_polyline(self, source: str, start: int or float = None, end: int or float = None,
                     point_sequence: PointSequence = None,
                     name: str or None = None, assign_unique_name: bool = True):
        polyline = PolyLine(source=source, start=start, end=end, point_sequence=point_sequence,
                            name=name)
        if point_sequence is None:
            for point in polyline.point_sequence().points():
                if point not in self.container()["points"].values():
                    self.add_pymead_obj_by_ref(point)
        return self.add_pymead_obj_by_ref(polyline, assign_unique_name=assign_unique_name)

    def split_polyline(self, polyline: PolyLine, split: int or float):
        new_polylines = polyline.split(split)

        # Remove the old polyline and its dependent points
        # if polyline.start is None and polyline.end is None:
        #     for point in polyline.point_sequence().points():
        #         self.remove_pymead_obj(point)
        #     polyline.point_sequence().points().clear()
        # else:
        #     self.remove_pymead_obj(polyline)
        self.remove_pymead_obj(polyline)
        for new_polyline in new_polylines:
            for point in new_polyline.point_sequence().points():
                if point not in self.container()["points"].values():
                    self.add_pymead_obj_by_ref(point)
            self.add_pymead_obj_by_ref(new_polyline)

    def add_desvar(self, value: float, name: str, lower: float or None = None, upper: float or None = None,
                   unit_type: str or None = None, assign_unique_name: bool = True, point: Point = None,
                   root: Point = None, rotation_handle: Point = None, enabled: bool = True, equation_str: str = None):
        """
        Directly adds a design variable value to the geometry collection.

        Parameters
        ==========
        value: float
            Value of the design variable

        name: str
            Name of the design variable (might be overridden when adding to the 'desvar' sub-container).

        lower: float or None
            Lower bound for the design variable. If ``None``, a reasonable value will be chosen automatically.
            Default: ``None``.

        upper: float or None.
            Upper bound for the design variable. If ``None``, a reasonable value will be chosen automatically.
            Default: ``None``.

        unit_type: str or None:
            The unit type of design variable to create. Default: ``None``.

        Returns
        =======
        DesVar
            The generated design variable
        """
        kwargs = dict(value=value, name=name, lower=lower, upper=upper, setting_from_geo_col=True, point=point,
                      root=root, rotation_handle=rotation_handle, enabled=enabled, equation_str=equation_str)
        if unit_type is None:
            desvar = DesVar(**kwargs)
        elif unit_type == "length":
            desvar = LengthDesVar(**kwargs)
        elif unit_type == "angle":
            desvar = AngleDesVar(**kwargs)
        else:
            raise ValueError(f"unit_type must be None, 'length', or 'angle'. Found type: {type(unit_type)}")

        return self.add_pymead_obj_by_ref(desvar, assign_unique_name=assign_unique_name)

    @staticmethod
    def replace_geo_objs(tool: Param or DesVar, target: Param or DesVar):
        """
        This static method is used to make sure that in param/desvar promotion/demotion, any references to geometric
        objects get replaced.

        Parameters
        ==========
        tool: Param or DesVar
            Object to be removed, and all geometric object references replaced with the target

        target: Param or DesVar
            Object to add
        """
        for geo_obj in tool.geo_objs:
            if isinstance(geo_obj, Point):
                if geo_obj.x() is tool:
                    geo_obj.set_x(target)
                elif geo_obj.y() is tool:
                    geo_obj.set_y(target)

    def promote_param_to_desvar(self, param: Param or str, lower: float or None = None, upper: float or None = None):
        """
        Promotes a parameter to a design variable by adding bounds. The ``Param`` will be removed from the 'params'
        sub-container, and the corresponding ``DesVar`` will be added to the 'desvar' sub-container.

        Parameters
        ==========
        param: Param or str
            Parameter to promote. If ``str``, the ``Param`` will be identified by looking in the 'params' sub-container.

        lower: float or None
            Lower bound for the design variable. If ``None``, a reasonable value will be chosen automatically.
            Default: ``None``.

        upper: float or none.
            Upper bound for the design variable. If ``None``, a reasonable value will be chosen automatically.
            Default: ``None``.

        Returns
        =======
        DesVar
            The generated design variable.
        """
        param = param if isinstance(param, Param) else self.container()["params"][param]

        if isinstance(param, LengthParam):
            unit_type = "length"
        elif isinstance(param, AngleParam):
            unit_type = "angle"
        else:
            unit_type = None

        desvar = self.add_desvar(value=param.value(), name=param.name(), lower=lower, upper=upper, unit_type=unit_type,
                                 point=copy(param.point), root=param.root,
                                 rotation_handle=param.rotation_handle, assign_unique_name=False)

        # Replace the corresponding x() or y() in parameter with the new design variable
        self.replace_geo_objs(tool=param, target=desvar)

        # Make a copy of the geometry object reference lists in the new design variable
        desvar.geo_objs = param.geo_objs.copy()

        # Copy constraint information
        desvar.geo_cons = param.geo_cons
        desvar.gcs = self.gcs
        for constraint in desvar.geo_cons:
            constraint.set_param(desvar)
        # desvar.gcs.constraint_params[self.gcs.constraint_params.index(param)] = desvar

        # Remove the parameter
        self.remove_pymead_obj(param, promotion_demotion=True)

        return desvar

    def demote_desvar_to_param(self, desvar: DesVar):
        """
        Demotes a design variable to a parameter by removing the bounds. The ``DesVar`` will be removed from the
        'desvar' sub-container, and the corresponding ``Param`` will be added to the 'params' sub-container.

        Parameters
        ==========
        desvar: DesVar
            Parameter to promote. If ``str``, the ``Param`` will be identified by looking in the 'params' sub-container.

        Returns
        =======
        Param
            The generated parameter
        """
        if isinstance(desvar, LengthDesVar):
            unit_type = "length"
        elif isinstance(desvar, AngleDesVar):
            unit_type = "angle"
        else:
            unit_type = None

        param = self.add_param(value=desvar.value(), name=desvar.name(), unit_type=unit_type, point=copy(desvar.point),
                               root=desvar.root, rotation_handle=desvar.rotation_handle, assign_unique_name=False)

        # Replace the corresponding x() or y() in parameter with the new parameter
        self.replace_geo_objs(tool=desvar, target=param)

        # Make a copy of the geometry object reference list in the new parameter
        param.geo_objs = desvar.geo_objs.copy()

        # Copy constraint information
        param.geo_cons = desvar.geo_cons
        param.gcs = self.gcs
        for constraint in param.geo_cons:
            constraint.set_param(param)
        # param.gcs.constraint_params[self.gcs.constraint_params.index(desvar)] = param

        # Remove the design variable
        self.remove_pymead_obj(desvar, promotion_demotion=True)

        return param

    def expose_point_xy(self, point: Point):
        self.add_pymead_obj_by_ref(point.x(), assign_unique_name=False)
        self.add_pymead_obj_by_ref(point.y(), assign_unique_name=False)

    def cover_point_xy(self, point: Point):
        self.remove_pymead_obj(point.x())
        self.remove_pymead_obj(point.y())

    def extract_design_variable_values(self, bounds_normalized: bool = False):
        """
        Extracts the values of the design variables in the geometry collection 'desvar' container.

        Parameters
        ==========
        bounds_normalized: bool
            Whether to normalize the design variable values by the bounds before extraction. Default: ``False``.

        Returns
        =======
        typing.List[float]
            List of design variable values, normalized the bounds if specified.
        """
        return [dv.value(bounds_normalized=bounds_normalized) for dv in self.container()["desvar"].values()]

    def assign_design_variable_values(self, dv_values: list, bounds_normalized: bool = False):
        """
        Assigns a list or array of design variable values, possibly normalized by the bounds, to the design variables
        in the geometry collection 'desvar' container.

        Parameters
        ==========
        dv_values: typing.Union[typing.Iterable, typing.Sized]
            List or array of design variable values. Must be equal in length to the number of design variables in the
            'desvar' container. Note that the value assignment will not necessarily take place in the same order as
            the order of the design variables in the GUI, but rather in the order they are present in the underlying
            dictionary.

        bounds_normalized: bool
            Whether the specified ``dv_values`` are normalized by the design variable bounds. Default: ``False``.
        """
        # First check that the length of the input vector and the number of design variables are equal
        dv_val_len = len(dv_values)
        num_dv = len(self.container()["desvar"])
        if dv_val_len != num_dv:
            raise ValueError(f"Length of design variable values to assign ({dv_val_len}) is different than the "
                             f"number of design variables ({num_dv})")

        # Set the values
        for dv, dv_value in zip(self.container()["desvar"].values(), dv_values):
            dv.set_value(dv_value, bounds_normalized=bounds_normalized)

    def alphabetical_sub_container_key_list(self, sub_container: str):
        """
        This method sorts a sub-container with the following rules:

        - Text case does not matter
        - Any consecutive numerical strings should appear in descending order
        - For parameters with an associated index, the parameter with an implied index of 1 should appear first,
          provided it exists. For example, ``"Point-2.x"`` should come after ``"Point.x"``, and ``"myParam-5"``
          should come after ``"myParam"``.

        Parameters
        ==========
        sub_container: str
            Sub-container from the geometry collection to sort alphabetically

        Returns
        =======
        typing.List[str]
            The alphabetically sorted list
        """
        original_list = [k for k in self.container()[sub_container].keys()]
        multiples = []
        for idx, k in enumerate(original_list):
            if "-" not in k:
                continue
            for sub_dot in k.split("."):
                if "-" not in sub_dot:
                    continue
                multiple = sub_dot.split("-")[0]
                if multiple not in multiples:
                    multiples.append(multiple)

        modified_list = []
        for idx, k in enumerate(original_list):
            if all([m not in k for m in multiples]):
                modified_list.append(k)
                continue
            k_split = k.split(".")
            for sub_idx, sub_dot in enumerate(k_split):
                if "-" in sub_dot or all([m != sub_dot for m in multiples]):
                    if sub_idx == 0:
                        modified_list.append(sub_dot)
                    else:
                        modified_list[-1] += sub_dot
                    if sub_idx < len(k_split) - 1:
                        modified_list[-1] += "."
                    continue
                modified_list.append(sub_dot + "-1")
                if sub_idx < len(k_split) - 1:
                    modified_list[-1] += "."

        sorted_list = sorted(modified_list, key=lambda k: [int(c) if c.isdigit() else c.lower()
                                                           for c in re.split("([0-9]+)", k)])
        for idx, ks in enumerate(sorted_list):
            if "-1." in ks:
                ks = ks.replace("-1.", ".")
            if ks[-2:] == "-1":
                ks = ks[:-2]
            sorted_list[idx] = ks

        return sorted_list

    def add_airfoil(self,
                    leading_edge: Point,
                    trailing_edge: Point,
                    upper_surf_end: Point or None = None,
                    lower_surf_end: Point or None = None,
                    name: str or None = None, assign_unique_name: bool = True):
        airfoil = Airfoil(leading_edge=leading_edge, trailing_edge=trailing_edge, upper_surf_end=upper_surf_end,
                          lower_surf_end=lower_surf_end, name=name)

        return self.add_pymead_obj_by_ref(airfoil, assign_unique_name=assign_unique_name)

    def add_mea(self, airfoils: typing.List[Airfoil], name: str or None = None, assign_unique_name: bool = True):
        mea = MEA(airfoils=airfoils, name=name)

        return self.add_pymead_obj_by_ref(mea, assign_unique_name=assign_unique_name)

    def add_constraint(self, constraint: GeoCon, assign_unique_name: bool = True, **constraint_kwargs):
        self.gcs.check_constraint_for_duplicates(constraint)
        self.add_pymead_obj_by_ref(constraint, assign_unique_name=assign_unique_name)
        if (constraint.param() is not None and constraint.param() not in self.container()["params"].values() and
                constraint.param() not in self.container()["desvar"].values()):
            self.add_pymead_obj_by_ref(constraint.param())
        try:
            self.gcs.add_constraint(constraint)
            if isinstance(constraint, AntiParallel3Constraint):
                if any([isinstance(curve, PolyLine) for curve in constraint.p1.curves]) and not constraint.p1.root:
                    self.gcs.move_root(constraint.p1)
                elif any([isinstance(curve, PolyLine) for curve in constraint.p3.curves]) and not constraint.p3.root:
                    self.gcs.move_root(constraint.p3)

            if (isinstance(constraint, AntiParallel3Constraint) or isinstance(constraint, Perp3Constraint) or
                    isinstance(constraint, RelAngle3Constraint) or isinstance(constraint, DistanceConstraint)):
                points_solved = self.gcs.solve(constraint)
                self.gcs.update_canvas_items(points_solved)
        except ValueError as e:
            self.remove_pymead_obj(constraint)
            self.clear_selected_objects()
            if self.gui_obj is not None:
                # self.gui_obj.showColoredMessage("Constraint cluster is over-constrained. Removing constraint...",
                #                                 4000, "#eb4034")
                self.gui_obj.disp_message_box(str(e), message_mode="error")
            return
        return constraint

    def equate_constraints(self, constraint1: GeoCon, constraint2: GeoCon):
        if constraint1.__class__.__name__ != constraint2.__class__.__name__:
            raise ValueError("Constraints must be of the same type to equate")

        self.remove_pymead_obj(constraint2.param(), equating_constraints=True)
        constraint2.set_param(constraint1.param())
        constraint1.param().geo_cons.append(constraint2)

        # Manually trigger an update by setting the value to the current value
        constraint1.param().set_value(constraint1.param().value())

    def get_dict_rep(self):
        dict_rep = {k_outer: {k: v.get_dict_rep() for k, v in self.container()[k_outer].items()}
                    for k_outer in self.container().keys()}
        dict_rep["metadata"] = self.get_metadata()
        return dict_rep

    @staticmethod
    def get_metadata():
        return {
            "pymead_version": __version__,
            "save_datetime": str(datetime.datetime.now()),
            "length_unit": UNITS.current_length_unit(),
            "angle_unit": UNITS.current_angle_unit(),
            "area_unit": UNITS.current_area_unit()
        }

    @classmethod
    def set_from_dict_rep(cls, d: dict, canvas=None, tree=None, gui_obj=None):
        geo_col = cls(gui_obj=gui_obj)
        geo_col.canvas = canvas
        geo_col.tree = tree
        for name, point_dict in d["points"].items():
            geo_col.add_point(**point_dict, name=name, assign_unique_name=False)
        for name, desvar_dict in d["desvar"].items():
            if ".x" in name:
                point = geo_col.container()["points"][name.split(".")[0]]
                geo_col.add_pymead_obj_by_ref(point.x(), assign_unique_name=False)
                geo_col.promote_param_to_desvar(point.x(), lower=desvar_dict["lower"], upper=desvar_dict["upper"])
            elif ".y" in name:
                point = geo_col.container()["points"][name.split(".")[0]]
                geo_col.add_pymead_obj_by_ref(point.y(), assign_unique_name=False)
                geo_col.promote_param_to_desvar(point.y(), lower=desvar_dict["lower"], upper=desvar_dict["upper"])
            else:
                root_name = desvar_dict.pop("root") if "root" in desvar_dict else None
                root = geo_col.container()["points"][root_name] if root_name is not None else None
                rotation_handle_name = desvar_dict.pop("rotation_handle") if "rotation_handle" in desvar_dict else None
                rotation_handle = geo_col.container()["points"][rotation_handle_name] \
                    if rotation_handle_name is not None else None
                geo_col.add_desvar(**desvar_dict, name=name, assign_unique_name=False, rotation_handle=rotation_handle,
                                   root=root)
        for name, param_dict in d["params"].items():
            if ".x" in name:
                point = geo_col.container()["points"][name.split(".")[0]]
                geo_col.add_pymead_obj_by_ref(point.x(), assign_unique_name=False)
            elif ".y" in name:
                point = geo_col.container()["points"][name.split(".")[0]]
                geo_col.add_pymead_obj_by_ref(point.y(), assign_unique_name=False)
            else:
                root_name = param_dict.pop("root") if "root" in param_dict else None
                root = geo_col.container()["points"][root_name] if root_name is not None else None
                rotation_handle_name = param_dict.pop("rotation_handle") if "rotation_handle" in param_dict else None
                rotation_handle = geo_col.container()["points"][rotation_handle_name] \
                    if rotation_handle_name is not None else None
                geo_col.add_param(**param_dict, name=name, assign_unique_name=False, rotation_handle=rotation_handle,
                                  root=root)
        for name, line_dict in d["lines"].items():
            geo_col.add_line(point_sequence=PointSequence(
                points=[geo_col.container()["points"][k] for k in line_dict["points"]]),
                name=name, assign_unique_name=False
            )
        if "polylines" in d:
            for name, polyline_dict in d["polylines"].items():
                geo_col.add_polyline(point_sequence=PointSequence(
                    points=[geo_col.container()["points"][k] for k in polyline_dict["points"]]),
                    source=polyline_dict["source"], start=polyline_dict["start"], end=polyline_dict["end"], name=name,
                    assign_unique_name=False)
        for name, bezier_dict in d["bezier"].items():
            geo_col.add_bezier(point_sequence=PointSequence(
                points=[geo_col.container()["points"][k] for k in bezier_dict["points"]]),
                name=name, assign_unique_name=False
            )

        constraints_added = []
        for name, geocon_dict in d["geocon"].items():
            for k, v in geocon_dict.items():
                if v in d["points"].keys():
                    geocon_dict[k] = geo_col.container()["points"][v]
                elif v in d["params"].keys():
                    geocon_dict[k] = geo_col.container()["params"][v]
                elif v in d["desvar"].keys():
                    geocon_dict[k] = geo_col.container()["desvar"][v]
                elif v in d["lines"].keys():
                    geocon_dict[k] = geo_col.container()["lines"][v]
                elif "polylines" in d and v in d["polylines"].keys():
                    geocon_dict[k] = geo_col.container()["polylines"][v]
                elif v in d["bezier"].keys():
                    geocon_dict[k] = geo_col.container()["bezier"][v]
                else:
                    pass
            constraint_type = geocon_dict.pop("constraint_type")
            constraint = getattr(sys.modules[__name__], constraint_type)(**geocon_dict, name=name)
            geo_col.add_constraint(constraint=constraint, assign_unique_name=False, compile=False,
                                   solve_and_update=False)
            constraints_added.append(constraint)
        for name, airfoil_dict in d["airfoils"].items():
            geo_col.add_airfoil(leading_edge=geo_col.container()["points"][airfoil_dict["leading_edge"]],
                                trailing_edge=geo_col.container()["points"][airfoil_dict["trailing_edge"]],
                                upper_surf_end=geo_col.container()["points"][airfoil_dict["upper_surf_end"]],
                                lower_surf_end=geo_col.container()["points"][airfoil_dict["lower_surf_end"]],
                                name=name, assign_unique_name=False)
        for name, mea_dict in d["mea"].items():
            geo_col.add_mea(airfoils=[geo_col.container()["airfoils"][k] for k in mea_dict["airfoils"]],
                            name=name, assign_unique_name=False)
        return geo_col

    def switch_units(self, unit_type: str, old_unit: str, new_unit: str):

        if old_unit == new_unit:
            return

        if unit_type == "length":
            UNITS.set_current_length_unit(new_unit)
        elif unit_type == "angle":
            UNITS.set_current_angle_unit(new_unit)

        def switch_unit_for_param(p: Param):
            if isinstance(p, LengthParam) and unit_type == "length":
                p.set_unit(new_unit, old_unit)

            elif isinstance(p, AngleParam) and unit_type == "angle":
                p.set_unit(new_unit, old_unit)

        def switch_unit_for_point(p: Point):
            if unit_type != "length":
                return

            new_x = p.x().set_unit(new_unit, old_unit, modify_value=False)
            new_y = p.y().set_unit(new_unit, old_unit, modify_value=False)
            p.request_move(new_x, new_y)

        for param in self.container()["params"].values():
            if param.point is not None:
                continue
            switch_unit_for_param(param)

        for desvar in self.container()["desvar"].values():
            if desvar.point is not None:
                continue
            switch_unit_for_param(desvar)

        for point in self.container()["points"].values():
            switch_unit_for_point(point)

        if unit_type == "length" and self.canvas is not None:
            x_data_range, y_data_range = self.canvas.getPointRange()
            self.canvas.plot.getViewBox().setRange(xRange=x_data_range, yRange=y_data_range)
            self.canvas.plot.setLabel(axis="bottom", text=f"x [{UNITS.current_length_unit()}]")
            self.canvas.plot.setLabel(axis="left", text=f"y [{UNITS.current_length_unit()}]")

    def write_to_iges(self, base_dir: str, file_name: str, translation: typing.List[float] = None,
                      scaling: typing.List[float] = None, rotation: typing.List[float] = None,
                      transformation_order: str = None):
        """
        Writes all the Bézier curves in the geometry collection to an IGES file.

        Parameters
        ----------
        base_dir: str
            Directory where the IGES file will be stored

        file_name: str
            Name of the IGES file (should include the .igs extension)

        translation: typing.List[float]
            How to translate the curves from the X-Z plane (in the form [tx, ty, tz])

        scaling: typing.List[float]
            How to scale the curves from the X-Z plane (in the form [sx, sy, sz])

        rotation: typing.List[float]
            How to rotate the curves from the X-Z plane (in the form [rx, ry, rz])

        transformation_order: str
            Order in which the transformation takes place

        Returns
        -------
        str
            The full file path to the created IGES file
        """

        # Create the full file path
        full_file = os.path.join(base_dir, file_name)

        # Grab the control points for each Bézier curve in the geometry collection
        control_point_list = [curve.point_sequence().as_array() for curve in self.container()["bezier"].values()]

        # Create the transformation object
        translation = [0., 0., 0.] if translation is None else translation
        scaling = [1., 1., 1.] if scaling is None else scaling
        rotation = [0., 0., 0.] if rotation is None else rotation
        transformation_order = "rx,ry,rz,s,t" if transformation_order is None else transformation_order
        transform_3d = Transformation3D(tx=[translation[0]], ty=[translation[1]], tz=[translation[2]],
                                        sx=[scaling[0]], sy=[scaling[1]], sz=[scaling[2]],
                                        rx=[rotation[0]], ry=[rotation[1]], rz=[rotation[2]],
                                        rotation_units="deg",
                                        order=transformation_order)

        # Add a third dimension by inserting a vector of zeros to the matrix
        cp_list_3d = []
        for cp in control_point_list:
            cp = np.insert(cp, 1, 0, axis=1)
            cp_list_3d.append(cp)

        # Transform the points
        transformed_cp_list = [transform_3d.transform(P) for P in cp_list_3d]

        # Create the list of IGES Bezier objects
        bez_IGES_list = [BezierIGES(P) for P in transformed_cp_list]

        # Generate the IGES file
        iges_generator = IGESGenerator(bez_IGES_list)
        iges_generator.generate(full_file)

        return full_file

    def verify_constraints(self):
        for geo_con in self.container()["geocon"].values():
            assert geo_con.verify()

        return True

    def verify_point_movement(self):
        random.seed(1)
        for point in self.container()["points"].values():
            # Record the starting xy position of the point
            old_xy = [point.x().value(), point.y().value()]
            random_xy = [random.uniform(-5.0, 5.0), random.uniform(-5.0, 5.0)]
            point.request_move(random_xy[0], random_xy[1])

            self.verify_constraints()

            # Return the point to its original position
            point.request_move(old_xy[0], old_xy[1])

        return True

    def verify_desvar(self):
        for desvar in self.container()["desvar"].values():

            # Record the starting value of the design variable
            old_value = desvar.value()

            # To avoid issues with setting a length of 0, set the value close to 0 if the lower bound is 0
            if isinstance(desvar, LengthParam) and np.isclose(desvar.lower(), 0.0):
                desvar.set_value(0.0001)
            else:
                desvar.set_value(desvar.lower())

            self.verify_constraints()

            desvar.set_value(desvar.upper())

            self.verify_constraints()

            # Set the design variable to its original value
            desvar.set_value(old_value)

        return True

    def verify_params(self):
        random.seed(1)
        for param in self.container()["params"].values():
            # Record the starting value of the param
            old_value = param.value()

            if isinstance(param, LengthParam):
                param.set_value(random.uniform(0.0001, 10.0))
            elif isinstance(param, AngleParam):
                param.set_value(random.uniform(0.0, 2 * np.pi))
            else:
                param.set_value(random.uniform(-10.0, 10.0))

            self.verify_constraints()

            param.set_value(old_value)

        return True

    def verify_all(self):
        self.verify_constraints()
        self.verify_point_movement()
        self.verify_desvar()
        self.verify_params()

        return True



from PyQt5.QtWidgets import QGraphicsView
from PyQt5.QtCore import Qt


class CustomGraphicsView(QGraphicsView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setDragMode(self.ScrollHandDrag)

    def wheelEvent(self, event) -> None:
        if event.angleDelta().y() > 0:
            scale = 1.25
        else:
            scale = .8
        self.scale(scale, scale)



import numpy as np


def translate_matrix(mat: np.ndarray, dx, dy):
    if not mat.shape[1] == 2:
        raise ValueError("Input matrix must be N x 2")
    mat[:, 0] += dx
    mat[:, 1] += dy
    return mat


def scale_matrix(mat: np.ndarray, scale_factor):
    if not mat.shape[1] == 2:
        raise ValueError("Input matrix must be N x 2")
    mat *= scale_factor
    return mat


def rotate_matrix(mat: np.ndarray, theta):
    if not mat.shape[1] == 2:
        raise ValueError("Input matrix must be N x 2")
    xy = mat.T
    rotation_mat = np.array([[np.cos(theta), -np.sin(theta)],
                             [np.sin(theta), np.cos(theta)]])
    mat = (rotation_mat @ xy).T
    return mat


def transform_matrix(mat: np.ndarray, dx, dy, theta, scale_factor, transformation_order: list):
    command_dict = {'translate': translate_matrix, 'rotate': rotate_matrix, 'scale': scale_matrix}
    argument_dict = {'translate': {'dx': dx, 'dy': dy}, 'rotate': {'theta': theta},
                     'scale': {'scale_factor': scale_factor}}
    for command in transformation_order:
        mat = command_dict[command](mat, **argument_dict[command])
    return mat



import typing
from copy import deepcopy

import numpy as np
import shapely.errors
from scipy.optimize import minimize
from shapely.geometry import Polygon

from pymead.core.geometry_collection import GeometryCollection
from pymead.utils.get_airfoil import extract_data_from_airfoiltools


def airfoil_symmetric_area_difference(parameters: list, geo_col: GeometryCollection, target_airfoil: str,
                                      airfoil_to_match_xy: np.ndarray):
    r"""
    This method uses the ``shapely`` package to convert the parametrized airfoil and the "discrete" airfoil to
    `Polygon <https://shapely.readthedocs.io/en/stable/manual.html#Polygon>`_
    objects and calculate the boolean
    `symmetric difference <https://shapely.readthedocs.io/en/stable/manual.html#object.symmetric_difference>`_
    (a similarity metric) between the two airfoils.

    Parameters
    ----------
    parameters: list
        A list of parameter values used to override the design variable values found in the geometry collection.

    geo_col: GeometryCollection
        Geometry collection from which the ``Airfoil`` is selected by the ``target_airfoil`` name and where the
        design variables are stored. During the optimization, any values in the design variable sub-container will
        be updated to produce an airfoil geometry that closely matches the airfoil coordinates specified by
        ``airfoil_to_match``.

    target_airfoil: str
        Airfoil from the geometry collection to match. For example, use ``"Airfoil-1"`` to match the airfoil
        with the same name found in the airfoil sub-container of the geometry collection. Only one airfoil may be
        matched at a time.

    airfoil_to_match_xy: np.ndarray
        Set of airfoil :math:`xy`-coordinates to be matched

    Returns
    -------
    float
        The boolean symmetric area difference between the parametrized airfoil and the discrete airfoil
    """

    # Override airfoil parameters with supplied sequence of parameters
    geo_col.assign_design_variable_values(parameters, bounds_normalized=True)
    airfoil = geo_col.container()["airfoils"][target_airfoil]
    coords = airfoil.get_coords_selig_format()

    # Calculate the boolean symmetric area difference. If there is a topological error, such as a self-intersection,
    # which prevents Polygon.area() from running, then make the symmetric area difference a large value to discourage
    # the optimizer from continuing in that direction.
    try:
        airfoil_shapely_points = list(map(tuple, coords))
        airfoil_polygon = Polygon(airfoil_shapely_points)
        airfoil_to_match_shapely_points = list(map(tuple, airfoil_to_match_xy))
        airfoil_to_match_polygon = Polygon(airfoil_to_match_shapely_points)
        symmetric_area_difference_polygon = airfoil_polygon.symmetric_difference(airfoil_to_match_polygon)
        symmetric_area_difference = symmetric_area_difference_polygon.area
        print(symmetric_area_difference)
    except shapely.errors.TopologicalError:
        symmetric_area_difference = 1  # Set the boolean symmetric area difference to a large value

    return symmetric_area_difference


def match_airfoil(geo_col: GeometryCollection, target_airfoil: str, airfoil_to_match: str or np.ndarray,
                  repair: typing.Callable or None = None):
    r"""
    This method uses the `sequential least-squares programming
    <https://docs.scipy.org/doc/scipy/reference/optimize.minimize-slsqp.html>`_ optimization scheme to minimize the
    boolean symmetric area difference between the input ``Airfoil`` and a set of "discrete" airfoil
    coordinates

    Parameters
    ----------
    geo_col: GeometryCollection
        Geometry collection from which the ``Airfoil`` is selected by the ``target_airfoil`` name and where the
        design variables are stored. During the optimization, any values in the design variable sub-container will
        be updated to produce an airfoil geometry that closely matches the airfoil coordinates specified by
        ``airfoil_to_match``.

    target_airfoil: str
        Airfoil from the geometry collection to match. For example, use ``"Airfoil-1"`` to match the airfoil
        with the same name found in the airfoil sub-container of the geometry collection. Only one airfoil may be
        matched at a time.

    airfoil_to_match: str or np.ndarray
        Set of airfoil :math:`xy`-coordinates to be matched, or a string representing the name
        of the airfoil to be fetched from `Airfoil Tools <http://airfoiltools.com/>`_.

    repair: typing.Callable or None
        An optional function that takes that makes modifications to the set of :math:`xy`-coordinates loaded from
        Airfoil Tools. This function should take exactly one input (the :math:`N \times 2` ``numpy.ndarray``
        representing the :math:`xy`-coordinates downloaded from Airfoil Tools) and return this array as the output.
        Default: ``None``

    Returns
    -------
    scipy.optimize.OptimizeResult
        `Results object <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html#scipy.optimize.OptimizeResult>`_
        returned by the optimizer
    """
    if isinstance(airfoil_to_match, str):
        airfoil_to_match_xy = extract_data_from_airfoiltools(airfoil_to_match, repair=repair)
    elif isinstance(airfoil_to_match, np.ndarray):
        airfoil_to_match_xy = airfoil_to_match
    else:
        raise TypeError(f'airfoil_to_match be of type str or np.ndarray, '
                        f'and type {type(airfoil_to_match)} was used')

    if target_airfoil not in geo_col.container()["airfoils"]:
        raise ValueError(f"Target airfoil {target_airfoil} not found in the specified geometry collection. Available"
                         f" airfoils in this geometry collection: "
                         f"{[k for k in geo_col.container()['airfoils'].keys()]}")

    if len(geo_col.container()["desvar"]) == 0:
        raise ValueError(f"No design variables were found in the geometry collection. Promote at least "
                         f"one parameter to a design variable to run an airfoil matching optimization.")

    geo_col_deepcopy = deepcopy(geo_col)
    initial_guess = np.array(geo_col.extract_design_variable_values())
    bounds = np.repeat(np.array([[0.0, 1.0]]), len(initial_guess), axis=0)
    res = minimize(
        airfoil_symmetric_area_difference,
        initial_guess,
        method="SLSQP",
        bounds=bounds,
        args=(geo_col_deepcopy, target_airfoil, airfoil_to_match_xy),
        options=dict(disp=True)
    )
    return res



import os
import typing

import matplotlib.pyplot as plt
import matplotlib.colors as mpl_colors
import numpy
from matplotlib import animation
from matplotlib.lines import Line2D
from matplotlib import ticker as mticker
from cycler import cycler
from matplotlib.patches import Polygon, Patch
import numpy as np
from copy import deepcopy
import PIL
from pymoo.decomposition.asf import ASF

from pymead.optimization.objectives_and_constraints import Objective
from pymead.utils.read_write_files import load_data, save_data
from pymead.optimization.pop_chrom import Chromosome, Population
from pymead.post.mses_field import generate_field_matplotlib
from pymead.analysis.read_aero_data import read_bl_data_from_mses
from pymead.core.geometry_collection import GeometryCollection
from pymead.utils.transformations import rotate_matrix
from pymead.post.fonts_and_colors import ILLINI_ORANGE, ILLINI_BLUE, font
from pymead.post.plot_formatters import format_axis_scientific, show_save_fig, legend_entry_flip
from pymead import DATA_DIR


ylabel = {
            "Cd": r"Design $C_d$ (counts)",
            "Cl": r"Design $C_l$",
            "Cdh": r"Design $C_{d_h}$ (counts)",
            "Cdw": r"Design $C_{d_w}$ (counts)",
            "alf": r"Design $\alpha$ (deg)",
            "Cdv": r"Design $C_{d_v}$ (counts)",
            "Cdp": r"Design $C_{d_p}$ (counts)",
            "Cdf": r"Design $C_{d_f}$ (counts)",
            "Cm": r"Design $C_m$",
            "CPK": r"Design $C_{P_K}$",
            "capSS": r"Supersonic Capture Streamtube",
            "sf": r"Design $C_{F_{\parallel V_\infty}}$ (counts)",
            "sfmh": r"Design $C_{F_{\parallel V_\infty}} - C_{d_h}$ (counts)",
            "Edot": r"Design $C_{\dot{\mathcal{E}}}$",
            "diss_surf": r"Design $C_{\Phi_{\mathrm{surf}}}$",
            "diss_shock": r"Design $C_{\Phi_{\mathrm{shock}}}$",
        }

bl_matplotlib_labels = {
    "Cp": r"$C_p$",
    "delta*": r"$\delta^*$",
    "theta": r"$\theta$",
    "theta*": r"$\theta^*$"
}


def field_display_text(index: int, display_var: str or typing.List[str], display_val: float or typing.List[float]):
    title = {
        "Cd": r"C_{F_{\parallel V_\infty}}",
        "CPK": r"C_{P_K}"
    }

    def output_display_text(disp_var: str, disp_val: float):
        disp_text = {
            "Cd": fr"${title['Cd']} = {disp_val:.3f}$",
            "CPK": fr"${title['CPK']} = {disp_val:.3f}$"
        }
        if disp_var not in disp_text.keys():
            raise ValueError(f"Failed to find display_var {disp_var} in display text keys ({disp_text.keys()})")
        return disp_text[disp_var]

    if isinstance(display_var, str) and isinstance(display_val, float):
        return f"Gen. {index}: " + output_display_text(display_var, display_val)
    elif isinstance(display_var, list) and isinstance(display_val, list):
        texts_to_join = []
        for dvar, dval in zip(display_var, display_val):
            texts_to_join.append(output_display_text(dvar, dval))
        return f"Gen. {index}: " + ", ".join(texts_to_join)
    else:
        raise TypeError("display_var and display_val must be str and float or List[str] and List[float]")


class PostProcess:
    def __init__(self, analysis_dir: str, image_dir: str, post_process_force_file: str, jmea_file: str,
                 modify_param_dict_func: typing.Callable = None, underwing: bool = False,
                 modify_param_dict_kwargs: dict = None, weights: np.ndarray = None,
                 settings_json_file: str = None):
        if modify_param_dict_kwargs is None:
            modify_param_dict_kwargs = {}
        self.cbar = None
        self.coord_plot_handles = []
        self.text_handles = []
        self.solid_handles = []
        self.analysis_dir = analysis_dir
        self.image_dir = image_dir
        if not os.path.exists(self.image_dir):
            os.mkdir(self.image_dir)
        self.post_process_force_file = post_process_force_file
        self.jmea_file = jmea_file
        self.geo_col_dict = load_data(jmea_file)
        self.geo_col_dict['airfoil_graphs_active'] = False
        self.underwing = underwing
        self.param_dict = load_data(os.path.join(self.analysis_dir, 'param_dict.json'))
        self.pop_size = deepcopy(self.param_dict["population_size"])
        print(f"{self.pop_size = }")
        self.param_dict_original = deepcopy(self.param_dict)
        self.modify_param_dict_func = modify_param_dict_func
        if self.modify_param_dict_func is not None:
            self.modify_param_dict(**modify_param_dict_kwargs)
        self.chromosomes = []
        self.multi_objective = self.determine_multi_objective()
        self.weights = weights if weights is not None else np.array([[0.5, 0.5]])
        self.settings_json_file = settings_json_file
        self.settings = None
        if self.settings_json_file is not None:
            self.settings = load_data(self.settings_json_file)

    def determine_multi_objective(self):
        alg1 = load_data(os.path.join(self.analysis_dir, f"algorithm_gen_1.pkl"))
        if alg1.opt.get("F").shape[1] > 1:
            return True
        return False

    def modify_param_dict(self, *args, **kwargs):
        self.modify_param_dict_func(self.param_dict, *args, **kwargs)

    def get_max_gen(self):
        max_gen = 1
        while True:
            if os.path.exists(os.path.join(self.analysis_dir, f"algorithm_gen_{max_gen}.pkl")):
                max_gen += 1
            else:
                return max_gen - 1

    def get_X(self, alg_file: str, weight_idx: int = 0):
        alg = load_data(alg_file)
        if alg.opt is None:
            return None
        else:
            if self.multi_objective:
                decomp = ASF()
                F = alg.opt.get("F")
                # I = decomp.do(F, self.weights[weight_idx]).argmin()
                F2 = deepcopy(F)
                # Normalize the objective function values before calculating the decomposition
                max_f1 = np.max(F2[:, 0])
                max_f2 = np.max(F2[:, 1])
                F2[:, 0] *= 1.0 / max_f1
                F2[:, 1] *= 1.0 / max_f2
                I2 = decomp.do(F2, self.weights[weight_idx]).argmin()
                return alg.opt.get("X")[I2]
            else:
                return alg.opt.get("X").flatten()

    def get_F(self, alg_file: str or None, weight_idx: int = 0):
        alg = None
        if alg_file is not None:
            alg = load_data(alg_file)
        if alg is None:
            baseline_file = os.path.join(self.analysis_dir, "J_baseline.dat")
            if os.path.exists(baseline_file):
                F = np.loadtxt(baseline_file)
                print(f"1, {F = }")
            else:
                F = self.get_baseline_obj_fun_values()
                print(f"2, {F = }")
            return F
        else:
            if self.multi_objective:
                decomp = ASF()
                F = alg.opt.get("F")

                # Normalize the objective function values before calculating the decomposition
                max_f1 = np.max(F[:, 0])
                max_f2 = np.max(F[:, 1])
                F[:, 0] *= 1.0 / max_f1
                F[:, 1] *= 1.0 / max_f2
                I_ = decomp.do(F, self.weights[weight_idx]).argmin()
                return alg.opt.get("F")[I_]
            else:
                return alg.opt.get("F").flatten()

    def get_full_gen_opt_F(self, alg_file: str or None):
        alg = None
        if alg_file is not None:
            alg = load_data(alg_file)
        if alg is None:
            baseline_file = os.path.join(self.analysis_dir, "J_baseline.dat")
            if os.path.exists(baseline_file):
                F = np.loadtxt(baseline_file)
            else:
                F = self.get_baseline_obj_fun_values()
            return F
        else:
            return alg.opt.get("F")

    def get_full_gen_pop_F(self, alg_file: str or None):
        alg = None
        if alg_file is not None:
            alg = load_data(alg_file)
        if alg is None:
            baseline_file = os.path.join(self.analysis_dir, "J_baseline.dat")
            if os.path.exists(baseline_file):
                F = np.loadtxt(baseline_file)
            else:
                F = self.get_baseline_obj_fun_values()
            return F
        else:
            return alg.pop.get("F")

    def get_mea(self, index: int, weight_idx: int = 0):
        if index != 0:
            X = self.get_X(os.path.join(self.analysis_dir, f"algorithm_gen_{index}.pkl"), weight_idx=weight_idx)
        else:
            X = None

        mea_object = GeometryCollection.set_from_dict_rep(self.geo_col_dict)

        if X is not None:
            mea_object.update_parameters(X)

        return mea_object

    def set_index(self, index: int or typing.Iterable = None):
        if isinstance(index, int):
            index = [index]
        elif index is None:
            index = np.arange(0, self.get_max_gen() + 1)
        return index

    def get_weight_str(self, weight_idx: int):
        weights = self.weights[weight_idx, :]
        if self.multi_objective:
            weight_str = "_w" + "".join([f"{int(100 * w)}-" for w in weights])[:-1]
        else:
            weight_str = ""
        return weight_str

    def get_baseline_obj_fun_values(self):
        if not os.path.exists(os.path.join(self.analysis_dir, 'analysis')):
            os.mkdir(os.path.join(self.analysis_dir, 'analysis'))

        param_set = deepcopy(self.param_dict_original)
        param_set['num_processors'] = 1
        param_set['population_size'] = 1
        param_set['mplot_settings']['flow_field'] = 2
        param_set['mses_settings']['timeout'] += 10.0
        param_set['mset_settings']['airfoil_analysis_dir'] = os.path.join(
            self.analysis_dir, 'analysis', f'analysis_0')
        param_set['mset_settings']['airfoil_coord_file_name'] = f'analysis_0'
        param_set['base_folder'] = os.path.join(self.analysis_dir, 'analysis')
        param_set['name'] = [f"analysis_0"]

        airfoil_name, mea_name = None, None
        if param_set["tool"] == "XFOIL":
            airfoil_name = param_set["xfoil_settings"]["airfoil"]
        elif param_set["tool"] == "MSES":
            mea_name = param_set["mset_settings"]["mea"]

        chromosome = Chromosome(param_dict=param_set, population_idx=0, geo_col_dict=deepcopy(self.geo_col_dict),
                                genes=None, generation=0, mea_name=mea_name, airfoil_name=airfoil_name)

        population = Population(param_dict=param_set, generation=0, parents=[chromosome],
                                verbose=True, skip_parent_assignment=False)

        population.eval_pop_fitness()
        population_forces = population.population[0].forces
        obj_fun_strings = self.settings["Genetic Algorithm"]["J"].split(",")
        J_baseline = []
        for J in obj_fun_strings:
            objective = Objective("")
            objective.set_func_str(J)
            objective.update(population_forces)
            J_baseline.append(objective.value)
        J_baseline = np.array(J_baseline)
        np.savetxt(os.path.join(self.analysis_dir, "J_baseline.dat"), J_baseline)
        return J_baseline

    def run_analysis(self, index: int or typing.Iterable = None, evaluate: bool = True, save_coords: bool = False,
                     save_control_points: bool = False, save_airfoil_state: bool = False):
        for weight_idx, weights in enumerate(self.weights):
            weight_str = self.get_weight_str(weight_idx)

            self.chromosomes = []

            index = self.set_index(index)

            X_list = [self.get_X(os.path.join(self.analysis_dir, f"algorithm_gen_{i}.pkl"), weight_idx=weight_idx)
                      if i != 0 else None for i in index]

            if not os.path.exists(os.path.join(self.analysis_dir, 'analysis')):
                os.mkdir(os.path.join(self.analysis_dir, 'analysis'))

            for idx, i in enumerate(index):
                param_set = deepcopy(self.param_dict)
                param_set['mset_settings']['airfoil_analysis_dir'] = os.path.join(
                    self.analysis_dir, 'analysis', f'analysis_{i}' + weight_str)
                param_set['mset_settings']['airfoil_coord_file_name'] = f'analysis_{i}' + weight_str
                param_set['base_folder'] = os.path.join(self.analysis_dir, 'analysis')
                param_set['name'] = [f"analysis_{j}" + weight_str for j in index]

                airfoil_name, mea_name = None, None
                if param_set["tool"] == "XFOIL":
                    airfoil_name = param_set["xfoil_settings"]["airfoil"]
                elif param_set["tool"] == "MSES":
                    mea_name = param_set["mset_settings"]["mea"]

                # parent_chromosomes.append(Chromosome(param_set=param_set, population_idx=s, mea=mea, X=X))
                X = X_list[idx]
                self.chromosomes.append(Chromosome(param_dict=param_set, population_idx=i,
                                                   geo_col_dict=deepcopy(self.geo_col_dict), genes=X,
                                                   generation=0, mea_name=mea_name, airfoil_name=airfoil_name))

            population = Population(param_dict=self.param_dict, generation=0, parents=self.chromosomes,
                                    verbose=True, skip_parent_assignment=False)

            if evaluate:
                population.eval_pop_fitness()
                population_forces = {k: [c.forces[k] if c.forces is not None else None for c in population.population]
                                     for k in population.population[0].forces.keys()}
                post_process_file_split = os.path.splitext(self.post_process_force_file)
                save_data(population_forces, post_process_file_split[0] + weight_str + post_process_file_split[1])
            else:
                population.generate_chromosomes_parallel()

            if save_coords:
                if not os.path.exists(os.path.join(self.analysis_dir, f'coords{weight_str}')):
                    os.mkdir(os.path.join(self.analysis_dir, f'coords{weight_str}'))
                for idx, c in enumerate(population.population):
                    save_data([a.tolist() for a in c.coords], os.path.join(self.analysis_dir, f'coords{weight_str}', f'coords_{index[idx]}.json'))
            if save_control_points:
                if not os.path.exists(os.path.join(self.analysis_dir, f'control_points{weight_str}')):
                    os.mkdir(os.path.join(self.analysis_dir, f'control_points{weight_str}'))
                for idx, c in enumerate(population.population):
                    control_points = [[curve.point_sequence().as_array().tolist() for curve in a.curves] for a in c.mea.airfoils]
                    save_data(control_points, os.path.join(self.analysis_dir,
                                                             f'control_points{weight_str}', f'control_points_{index[idx]}.json'))
            if save_airfoil_state:
                # if not os.path.exists(os.path.join(self.analysis_dir, f'airfoil_state{weight_str}')):
                #     os.mkdir(os.path.join(self.analysis_dir, f'airfoil_state{weight_str}'))
                # for idx, c in enumerate(population.population):
                #     save_data(c.airfoil_state, os.path.join(self.analysis_dir,
                #                                             f'airfoil_state{weight_str}', f'airfoil_state_{index[idx]}.json'))
                pass

    def generate_objective_function_plot(self, index: int or typing.Iterable = None):
        obj_fun_label_dict = {
            0: r"$J_P$",
            1: r"$J_F$"
        }
        save_name_dict = {
            0: "JP",
            1: "JF",
        }

        F_list = []
        for weight_idx, weights in enumerate(self.weights):
            index = self.set_index(index)
            F = np.array([self.get_F(os.path.join(self.analysis_dir, f"algorithm_gen_{i}.pkl"), weight_idx=weight_idx)
                          if i != 0 else self.get_F(alg_file=None, weight_idx=weight_idx) for i in index])
            F_list.append(F)
            print(f"{F = }")

        for obj_fun_idx in range(F_list[0].shape[1]):
            fig, axs = plt.subplots()

            prop_cycler = (cycler(color=['#004488', '#DDAA33', '#BB5566']))
            axs.set_prop_cycle(prop_cycler)

            for i, F in enumerate(F_list):
                weight_str = self.get_weight_str(i)
                label = None if self.weights.shape[0] < 2 else weight_str[1:]
                label_dict = {
                    "w0-100": f"POP{self.pop_size}L",
                    "w50-50": f"POP{self.pop_size}M",
                    "w100-0": f"POP{self.pop_size}R"
                }
                label = label_dict[label]

                # Determine what ydata to plot
                xdata = np.arange(F.shape[0])
                ydata = F[:, obj_fun_idx]
                xdata = np.delete(xdata, np.argwhere(ydata == 1.0e9))
                ydata = np.delete(ydata, np.argwhere(ydata == 1.0e9))

                # Plot the data
                axs.plot(xdata, ydata, label=label)

            fig.set_tight_layout('tight')
            # if obj_fun_idx == 0 and self.pop_size == 150:
            #     axs.set_ylim([0.165, 0.235])
            axs.set_xlabel("Generation", fontdict=font)
            axs.set_ylabel(obj_fun_label_dict[obj_fun_idx], fontdict=font)
            axs.grid("on", ls=":")
            if self.weights.shape[0] > 1:
                legend_font_dict = {k: v for k, v in font.items() if k != "color"}
                axs.legend(prop=legend_font_dict)
            format_axis_scientific(axs)
            show_save_fig(fig, save_base_dir=self.image_dir, file_name_stub=f'{save_name_dict[obj_fun_idx]}')

    def generate_aero_force_plot(self, var: str or list = None):
        if self.multi_objective:
            weight_str = self.get_weight_str(0)
            post_process_file_split = os.path.splitext(self.post_process_force_file)
            post_process_forces = load_data(post_process_file_split[0] + weight_str + post_process_file_split[1])
            if self.weights.shape[0] > 1:
                post_process_forces_list = [load_data(
                    post_process_file_split[0] + self.get_weight_str(i) + post_process_file_split[1])
                    for i in range(self.weights.shape[0])]
            else:
                post_process_forces_list = [post_process_forces]
        else:
            post_process_forces = load_data(self.post_process_force_file)
            post_process_forces_list = [post_process_forces]
        if isinstance(var, str):
            var = [var]
        if var is None:
            var = [k for k in post_process_forces.keys() if k not in ['converged', 'BL', 'errored_out', 'timed_out',
                                                                      'epma']]
            var.extend(["sf", "sfmh"])
        multiplier = {
            "Cd": 10000,
            "Cl": 1,
            "Cdh": 10000,
            "Cdw": 10000,
            "alf": 1,
            "Cdv": 10000,
            "Cdp": 10000,
            "Cdf": 10000,
            "Cm": 1,
            "CPK": 1,
            "capSS": 1,
            "sf": 10000,
            "sfmh": 10000,
            "diss_shock": 1,
            "diss_surf": 1,
            "Edot": 1,
        }
        for v in var:
            fig, axs = plt.subplots()

            prop_cycler = (cycler(color=['#004488', '#DDAA33', '#BB5566']))
            axs.set_prop_cycle(prop_cycler)

            for i, ppf in enumerate(post_process_forces_list):
                weight_str = self.get_weight_str(i)
                label = None if self.weights.shape[0] < 2 else weight_str[1:]
                label_dict = {
                    "w0-100": f"POP{self.pop_size}L",
                    "w50-50": f"POP{self.pop_size}M",
                    "w100-0": f"POP{self.pop_size}R"
                }
                label = label_dict[label]

                # Determine what ydata to plot
                if v in ppf.keys():
                    if isinstance(post_process_forces['Cd'][0], typing.Iterable):
                        ydata = [el[1] * multiplier[v] for el in ppf[v]]
                    else:
                        ydata = [el * multiplier[v] for el in ppf[v]]
                elif v == "sf":
                    if isinstance(post_process_forces['Cd'][0], typing.Iterable):
                        ydata = [el[1] * multiplier[v] for el in ppf["Cd"]]
                    else:
                        ydata = [el * multiplier[v] for el in ppf["Cd"]]
                elif v == "sfmh":
                    if isinstance(post_process_forces['Cd'][0], typing.Iterable):
                        ydata = [(el_Cd[1] - el_Cdh[1]) * multiplier[v] for el_Cd, el_Cdh in zip(ppf["Cd"], ppf["Cdh"])]
                    else:
                        ydata = [(el_Cd - el_Cdh) * multiplier[v] for el_Cd, el_Cdh in zip(ppf["Cd"], ppf["Cdh"])]
                else:
                    raise ValueError(f"Found invalid key {v}")

                # Plot the ydata
                axs.plot(ydata, label=label)

            fig.set_tight_layout('tight')
            axs.set_xlabel("Generation", fontdict=font)
            axs.set_ylabel(ylabel[v], fontdict=font)
            axs.grid("on", ls=":")
            if self.weights.shape[0] > 1:
                legend_font_dict = {k: v for k, v in font.items() if k != "color"}
                axs.legend(prop=legend_font_dict)
            format_axis_scientific(axs)
            show_save_fig(fig, save_base_dir=self.image_dir, file_name_stub=f'design_{v}')

    def pareto_front(self, main_axes_xy_lim: typing.List, inset_axes_xy_lim: typing.List, axes_location: typing.List,
                     opt_start: int = 5, opt_end: int = None, opt_num: int = 10):

        fig, axs = plt.subplots(figsize=(8, 4.8))

        def generate_plots(ax: plt.Axes, opt_end_):
            index = self.set_index(None)
            if opt_end_ is None:
                opt_end_ = index[-1]
            all_opt_F = [self.get_full_gen_opt_F(os.path.join(self.analysis_dir, f"algorithm_gen_{i}.pkl"))
                     if i != 0 else self.get_full_gen_opt_F(alg_file=None) for i in index]
            all_pop_F = [self.get_full_gen_pop_F(os.path.join(self.analysis_dir, f"algorithm_gen_{i}.pkl"))
                         if i != 0 else self.get_full_gen_pop_F(alg_file=None) for i in index]
            opt_range = np.linspace(opt_start, opt_end_, opt_num, endpoint=True).astype(int)
            cmap = plt.get_cmap("plasma", len(opt_range))
            cmap_dict = {v: i for i, v in enumerate(opt_range)}

            pop_labeled = False
            opt_labeled = False
            for i, F in enumerate(all_pop_F):
                if i != 0:
                    extra_kwargs = {}
                    if not pop_labeled:
                        extra_kwargs["label"] = "Individual"
                        pop_labeled = True
                    valid_rows = np.argwhere(F[:, 0] < 1.0).flatten()
                    ax.plot(F[valid_rows, 0], F[valid_rows, 1], ls="none", marker="s", mfc="lightgray", mec="lightgray", alpha=0.3,
                             markersize=2, **extra_kwargs)
            for i, F in enumerate(all_opt_F):
                if F.ndim > 1:
                    sorted_order = np.argsort(F[:, 0], axis=None)
                    F[:, 0] = F[sorted_order, 0]
                    F[:, 1] = F[sorted_order, 1]
                star_kwargs = dict(ls="--", marker="*")
                gray_circle_kwargs = dict(ls="none", marker="o", mfc="none", mec="gray", markersize=3)
                base_kwargs = dict(ls="none", marker="x", mfc="red", mec="red")
                if i == 0:
                    axs.plot(F[0], F[1], **base_kwargs, label="Baseline")
                else:
                    if i in opt_range:
                        ax.plot(F[:, 0], F[:, 1], **star_kwargs, zorder=100, label=f"Opt, Gen {i}",
                                 color=cmap(cmap_dict[i]),
                                 mfc=cmap(cmap_dict[i]),
                                 mec=cmap(cmap_dict[i])
                                 )
                    else:
                        extra_kwargs = {}
                        if not opt_labeled:
                            extra_kwargs["label"] = "Optimal"
                            opt_labeled = True
                        ax.plot(F[:, 0], F[:, 1], **gray_circle_kwargs, **extra_kwargs)
            # ax.set_xlim([0.058, 0.19])  # FOR POP50 input ga_opt_70
            ax.set_xlim(main_axes_xy_lim[:2])  # FOR POP50 consumption ga_opt_93
            ax.set_ylim(main_axes_xy_lim[2:])
            # ax.set_xlim([0.17, 0.26])  # FOR POP150 consumption ga_opt_92
            ax.set_xlabel(r"$J_P$", fontdict=font)
            ax.set_ylabel(r"$J_F$", fontdict=font)
            format_axis_scientific(ax)

        generate_plots(axs, opt_end)
        legend_font_dict = {k: v for k, v in font.items() if k != "color"}
        legend_font_dict["size"] = 12
        axs.legend(prop=legend_font_dict, loc="upper right", ncol=1)

        axs_inset = axs.inset_axes(axes_location)
        generate_plots(axs_inset, opt_end)
        x1, x2, y1, y2 = inset_axes_xy_lim[0], inset_axes_xy_lim[1], inset_axes_xy_lim[2], inset_axes_xy_lim[3]
        axs_inset.set_xlim(x1, x2)
        axs_inset.set_ylim(y1, y2)
        # x_labels = [item if idx % 2 else "" for idx, item in enumerate(axs_inset.get_xticklabels())]
        # y_labels = [item if not idx % 2 else "" for idx, item in enumerate(axs_inset.get_yticklabels())]
        # axs_inset.set_xticklabels(x_labels)
        # axs_inset.set_yticklabels(y_labels)
        axs_inset.xaxis.set_major_locator(mticker.MaxNLocator(2))
        axs_inset.yaxis.set_major_locator(mticker.MaxNLocator(2, prune="lower"))
        axs_inset.set_zorder(100)

        show_save_fig(fig, save_base_dir=self.image_dir, file_name_stub="pareto_front")

    def compare_geometries(self, index: list, plot_actuator_disk: bool = True,
                           rotate_x_axis_wind_direction: bool = False, rotate_zero_alf: bool = False):
        for weight_idx, weights in enumerate(self.weights):
            weight_str = self.get_weight_str(weight_idx)
            if len(index) != 2:
                raise ValueError(f"Comparison of only 2 geometries is supported. Current index list: {index}")
            fig, axs = plt.subplots(figsize=(12, 5))
            colors = [ILLINI_BLUE, ILLINI_ORANGE]
            save_filetypes = ['.svg', '.pdf']
            post_process_forces = None
            if rotate_x_axis_wind_direction:
                post_process_force_file = os.path.splitext(
                    self.post_process_force_file)[0] + weight_str + os.path.splitext(self.post_process_force_file)[-1]
                post_process_forces = load_data(post_process_force_file)
            for analysis_idx, i in enumerate(index):
                coords = load_data(os.path.join(self.analysis_dir, f'coords{weight_str}', f'coords_{i}.json'))
                alf = None
                if rotate_zero_alf:
                    dx = (coords[0][0][0] + coords[0][-1][0]) / 2
                    dy = (coords[0][0][1] + coords[0][-1][1]) / 2
                    alf = -np.arctan2(dy, dx)
                for airfoil in coords:
                    airfoil = np.array(airfoil)
                    if rotate_x_axis_wind_direction:
                        airfoil = rotate_matrix(airfoil, -np.deg2rad(post_process_forces["alf"][i][1]))
                    if rotate_zero_alf:
                        airfoil = rotate_matrix(airfoil, alf)
                    axs.plot(airfoil[:, 0], airfoil[:, 1], color=colors[analysis_idx], ls='dashed')
                if plot_actuator_disk:
                    control_points = load_data(os.path.join(self.analysis_dir,
                                                            f'control_points{weight_str}', f'control_points_{i}.json'))
                    # Index order: [airfoil][curve][point][x or y]
                    if self.underwing:
                        ad1_x = [control_points[1][2][0][0], control_points[0][3][0][0]]
                        ad1_y = [control_points[1][2][0][1], control_points[0][3][0][1]]
                        ad2_x = [control_points[2][1][0][0], control_points[1][4][0][0]]
                        ad2_y = [control_points[2][1][0][1], control_points[1][4][0][1]]
                    else:
                        ad1_x = [control_points[1][2][0][0], control_points[2][3][0][0]]
                        ad1_y = [control_points[1][2][0][1], control_points[2][3][0][1]]
                        ad2_x = [control_points[0][1][0][0], control_points[1][4][0][0]]
                        ad2_y = [control_points[0][1][0][1], control_points[1][4][0][1]]

                    ad1 = np.column_stack((np.array(ad1_x), np.array(ad1_y)))
                    ad2 = np.column_stack((np.array(ad2_x), np.array(ad2_y)))
                    if rotate_x_axis_wind_direction:
                        ad1 = rotate_matrix(ad1, -np.deg2rad(post_process_forces["alf"][i][1]))
                        ad2 = rotate_matrix(ad2, -np.deg2rad(post_process_forces["alf"][i][1]))
                    if rotate_zero_alf:
                        ad1 = rotate_matrix(ad1, alf)
                        ad2 = rotate_matrix(ad2, alf)
                    axs.plot(ad1[:, 0], ad1[:, 1], color=colors[analysis_idx], ls='dotted')
                    axs.plot(ad2[:, 0], ad2[:, 1], color=colors[analysis_idx], ls='dotted')

            legend_proxies = [Line2D([], [], color=c, ls='dashed') for c in colors]
            legend_names = ['initial', 'optimal']
            ncols = 1
            if plot_actuator_disk:
                legend_proxies.extend([Line2D([], [], color=c, ls='dotted') for c in colors])
                ncols += 1
                legend_names.extend(['initial AD', 'optimal AD'])

            if self.underwing:
                legend_loc = "lower left"
            else:
                legend_loc = "upper left"
            axs.legend(legend_proxies, legend_names, prop={'size': 18, 'family': font['family']},
                       fancybox=False, shadow=False, loc=legend_loc, ncol=ncols, frameon=False)
            axs.set_aspect('equal')
            axs.set_xlabel(r'$x/c_{main}$', fontdict=dict(family="serif", size=18))
            axs.set_ylabel(r'$y/c_{main}$', fontdict=dict(family="serif", size=18))
            axs.grid('on', ls=':')
            format_axis_scientific(axs)
            modifiers = ""
            if rotate_x_axis_wind_direction:
                modifiers += "_rotateWind"
            if rotate_zero_alf:
                modifiers += "_rotateZeroAlf"
            for ext in save_filetypes:
                fig.savefig(os.path.join(self.image_dir, f"geometry_{index[0]}_{index[1]}{modifiers}{weight_str}{ext}"),

                            bbox_inches="tight")

    def compare_camber_thickness(self, index: list, ls: tuple = (":", "--")):
        for weight_idx, weights in enumerate(self.weights):
            weight_str = self.get_weight_str(weight_idx)
            if len(index) != 2:
                raise ValueError(f"Comparison of only 2 geometries is supported. Current index list: {index}")

            fig1, ax1 = plt.subplots(figsize=(6, 6))
            fig2, ax2 = plt.subplots(figsize=(6, 6))
            figs = [fig1, fig2]
            axs = [ax1, ax2]
            f_name = ["camber", "thickness"]
            xc_array = np.linspace(0, 1, 250)

            legend_titles = {
                "A0": ("Main (Base)", "Main (Opt)"),
                "A1": ("Hub (Base)", "Hub (Opt)"),
                "A2": ("Nac (Base)", "Nac (Opt)"),
            }

            prop_cycler = (cycler(color=['#004488', '#DDAA33', '#BB5566']))

            for ax in axs:
                ax.set_prop_cycle(prop_cycler)

            for i, geometry_idx in enumerate(index):
                mea_object = self.get_mea(geometry_idx, weight_idx=weight_idx)
                for a_name, a in mea_object.airfoils.items():
                    label = legend_titles[a_name][i]
                    plot_kwargs = dict(label=label, ls=ls[i])
                    tc = a.compute_thickness_at_points(xc_array)
                    hc = a.compute_camber_at_points(xc_array)
                    axs[0].plot(xc_array, hc, **plot_kwargs)
                    axs[1].plot(xc_array, tc, **plot_kwargs)

            ncols = 3
            for ax in axs:
                handles, labels = ax.get_legend_handles_labels()
                ax.legend(legend_entry_flip(handles, ncols), legend_entry_flip(labels, ncols),
                          loc='upper center', bbox_to_anchor=(0.5, 1.2), ncol=ncols, fancybox=False, shadow=False, frameon=False,
                          prop=dict(family="serif", size=15))
                format_axis_scientific(ax)

            axis_label_font = dict(family="serif", size=18)
            axs[0].set_xlabel("Chord Fraction", fontdict=axis_label_font)
            axs[1].set_xlabel("Chord Fraction", fontdict=axis_label_font)
            axs[0].set_ylabel("Camber-to-Chord Ratio", fontdict=axis_label_font)
            axs[1].set_ylabel("Thickness-to-Chord Ratio", fontdict=axis_label_font)

            for i, fig in enumerate(figs):
                show_save_fig(fig, save_base_dir=self.image_dir, file_name_stub=f_name[i] + weight_str)

    def compare_airfoil_and_camber_to_known_airfoil(self, airfoil: str, index: int):
        fig1, ax1 = plt.subplots(figsize=(6, 3))
        fig2, ax2 = plt.subplots(figsize=(6, 3))
        figs = [fig1, fig2]
        axs = [ax1, ax2]
        f_name = [f"main_opt_{airfoil}_compare_coords", f"main_opt_{airfoil}_compare_camber"]
        xc_array = np.linspace(0, 1, 250)
        airfoil_coords = np.loadtxt(os.path.join(DATA_DIR, airfoil, f"{airfoil}.dat"))
        airfoil_camber = np.loadtxt(os.path.join(DATA_DIR, airfoil, f"{airfoil}_camber.dat"))
        legend_titles = ("Main (Opt)", airfoil)

        prop_cycler = (cycler(color=['#004488', '#DDAA33', '#BB5566']))

        for ax in axs:
            ax.set_prop_cycle(prop_cycler)

        mea_object = self.get_mea(index)
        a = mea_object.airfoils["A0"]
        hc = a.compute_camber_at_points(xc_array)
        opt_coords = mea_object.airfoils["A0"].get_coords(body_fixed_csys=True)
        axs[0].plot(opt_coords[:, 0], opt_coords[:, 1], label=legend_titles[0], ls="-.")
        axs[0].plot(airfoil_coords[:, 0], airfoil_coords[:, 1], label=legend_titles[1], ls="-.")
        axs[0].set_aspect("equal")
        axs[1].plot(xc_array, hc, label=legend_titles[0], ls="-.")
        axs[1].plot(airfoil_camber[:, 0], airfoil_camber[:, 1], label=legend_titles[1], ls="-.")
        axis_label_font = dict(family="serif", size=18)
        axs[0].set_xlabel(r"$x/c$", fontdict=axis_label_font)
        axs[1].set_xlabel(r"$x/c$", fontdict=axis_label_font)
        axs[0].set_ylabel(r"$y/c$", fontdict=axis_label_font)
        axs[1].set_ylabel(r"$h/c$", fontdict=axis_label_font)

        ncols = 2
        bbox = ((0.5, 1.8), (0.5, 1.22))
        for i, ax in enumerate(axs):
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(legend_entry_flip(handles, ncols), legend_entry_flip(labels, ncols),
                      loc='upper center', bbox_to_anchor=bbox[i], ncol=ncols, fancybox=False, shadow=False, frameon=False,
                      prop=dict(family="serif", size=15))
            format_axis_scientific(ax)

        for ax in axs:
            format_axis_scientific(ax)

        for i, fig in enumerate(figs):
            show_save_fig(fig, save_base_dir=self.image_dir, file_name_stub=f_name[i])

    def generate_BL_plot(self, var: str or typing.Iterable = None, index: int or typing.Iterable = None,
                         mode: str = 'standalone', legend_strs: tuple = ('upper', 'lower')):

        save_filetypes = ['.svg', '.pdf']
        pai_props = {
            0: dict(title='Main', xlim=[-0.1, 1.1], ylim=[2.2, -1.6], xlabel=r'$x/c_{main}$', ylabel=bl_matplotlib_labels[var]),
            1: dict(title='Hub', xlim=[0.75, 1.4], ylim=[2.2, -1.6], xlabel=r'$x/c_{main}$', ylabel=''),
            2: dict(title='Nacelle', xlim=[0.55, 1.1], ylim=[2.2, -1.6], xlabel=r'$x/c_{main}$', ylabel='')
        }

        valid_modes = ['compare', 'standalone']
        if mode not in valid_modes:
            raise ValueError(f"mode must be one of {valid_modes}")
        index = self.set_index(index)
        if isinstance(index, int):
            index = [index]
        if isinstance(var, str):
            var = [var]
        BL_data = {}

        fig, axs = None, None
        for v in var:
            for weight_idx, w in enumerate(self.weights):
                weight_str = self.get_weight_str(weight_idx)
                for analysis_idx, i in enumerate(index):
                    bl_file = os.path.join(self.analysis_dir, 'analysis',
                                           f'analysis_{i}' + weight_str, f'bl.analysis_{i}' + weight_str)
                    BL_data[i] = read_bl_data_from_mses(bl_file)
                    n_sides = len(BL_data[i])
                    if mode == 'compare' and analysis_idx == 0:
                        fig, axs = plt.subplots(nrows=1, ncols=int(n_sides / 2), figsize=(12, 3.3))
                    elif mode == 'standalone':
                        fig, axs = plt.subplots(nrows=1, ncols=int(n_sides / 2), figsize=(12, 3.3))
                    colors = [ILLINI_BLUE, ILLINI_ORANGE]
                    line_styles = ['dashed', 'dotted']
                    for side in range(n_sides):
                        color = None
                        if mode == 'compare':
                            color = colors[analysis_idx]
                        elif mode == 'standalone':
                            color = colors[0]
                        if self.underwing:
                            target_axs = int(side / 2)
                        else:
                            target_axs = 2 - int(side / 2)
                        axs[target_axs].plot(BL_data[i][side]['x'], BL_data[i][side][v], color=color,
                                             ls=line_styles[side % 2])
                    for k, v_ in pai_props.items():
                        axs[k].set_xlim(v_["xlim"])
                        axs[k].set_ylim(v_["ylim"])
                        axs[k].set_title(v_["title"], fontdict=font)
                        axs[k].set_xlabel(v_["xlabel"], fontdict=font)
                        axs[k].set_ylabel(v_["ylabel"], fontdict=font)
                        if v == "Cp":
                            axs[k].set_yticks([-1, 0, 1])
                        axs[k].grid('on', ls=':')
                    for k in range(3):
                        pos = axs[k].get_position()
                        axs[k].set_position([pos.x0, pos.y0 * 1.35, pos.width, pos.height * 0.83])
                    if v == 'Cp' and analysis_idx == 0:
                        for j in range(int(n_sides / 2)):
                            axs[j].invert_yaxis()

                    for ax in axs:
                        format_axis_scientific(ax)

                    # Set the legend
                    if mode == 'compare':
                        legend_proxies = [Line2D([], [], ls=ls, color=c)
                                          for ls, c in tuple(zip(line_styles * 2,
                                                                 [colors[0], colors[0], colors[1], colors[1]]))]
                        fig.legend(legend_proxies, legend_strs, fancybox=False, shadow=False, frameon=False, bbox_to_anchor=(0.5, 0.97),
                                   ncol=len(legend_proxies), loc='upper center', prop={'size': 14,
                                                                                       'family': font['family']})

                    if mode == 'standalone':
                        for ext in save_filetypes:
                            fig.savefig(os.path.join(self.image_dir, f'gen_{i}_{v}{weight_str}{ext}.svg'.replace('*', 'star')),
                                        bbox_inches="tight")
                if mode == 'compare':
                    for ext in save_filetypes:
                        fig.savefig(os.path.join(self.image_dir,
                                                 f'gen_{index[0]}_{index[1]}_{v}{weight_str}{ext}'.replace('*', 'star')),
                                    bbox_inches="tight")

    def generate_field_matplotlib(self, var: str, axs: plt.Axes or None, index: int, cmap_field: mpl_colors.Colormap or str,
                                  cmap_airfoil: mpl_colors.Colormap or str, shading: str = 'gouraud', vmin: float = None,
                                  vmax: float = None, weight_idx: int = 0):
        return generate_field_matplotlib(axs=axs,
                                         analysis_subdir=os.path.join(self.analysis_dir,
                                                                      'analysis', f'analysis_{index}{self.get_weight_str(weight_idx)}'),
                                         var=var, cmap_field=cmap_field, cmap_airfoil=cmap_airfoil, shading=shading,
                                         vmin=vmin, vmax=vmax)

    def generate_single_field(self, var: str, index: int, index_list, cmap_field: mpl_colors.Colormap or str,
                              cmap_airfoil: mpl_colors.Colormap or str, shading: str = 'gouraud', vmin: float = None,
                              vmax: float = None, image_extensions: tuple = ('.png', '.pdf'),
                              field_display_var: str or typing.List[str] = "Cd", weight_idx_list: list = None):
        for weight_idx, weights in enumerate(self.weights):

            if weight_idx_list is not None and weight_idx not in weight_idx_list:
                continue

            weight_str = self.get_weight_str(weight_idx)
            airfoil_color = 'black'
            flow_var_label_matplotlib = {'M': r'Mach Number',
                                         'Cp': r'Pressure Coefficient',
                                         'p': r'Static Pressure ($p/p_\infty$)',
                                         'rho': r'Density ($\rho/\rho_\infty$)',
                                         'u': r'Velocity-x ($u/V_\infty$)',
                                         'v': r'Velocity-y ($v/V_\infty$)',
                                         'q': r'Speed of Sound ($q/V_\infty$)',
                                         "Cpt": r"Total Pressure Over P_inf",
                                         "dCpt": r"Delta Total Pressure",
                                         "dCp": r"Delta Pressure Coefficient"}
            post_process_forces = load_data(
                os.path.splitext(
                    self.post_process_force_file)[0] + weight_str + os.path.splitext(self.post_process_force_file)[1])
            fig, axs = plt.subplots(figsize=(10, 5))
            # plt.subplots_adjust(left=0.08, bottom=-0.1, right=1.05, top=1.2, wspace=0.0, hspace=0.0)
            quad = generate_field_matplotlib(axs=axs,
                                      analysis_subdir=os.path.join(self.analysis_dir,
                                                                   'analysis', f'analysis_{index}{weight_str}'),
                                      var=var, cmap_field=cmap_field, cmap_airfoil=cmap_airfoil, shading=shading,
                                      vmin=vmin, vmax=vmax)

            # Plot the airfoils
            coords = load_data(os.path.join(self.analysis_dir, f'coords{weight_str}', f'coords_{index}.json'))
            for airfoil in coords:
                airfoil = np.array(airfoil)
                axs.plot(airfoil[:, 0], airfoil[:, 1], color=airfoil_color)
                polygon = Polygon(airfoil, closed=False, color="#000000AA")
                axs.add_patch(polygon)

            # Get location of index in gen_list
            if isinstance(index_list, list):
                gen_index = index_list.index(index)
            elif isinstance(index_list, numpy.ndarray):
                gen_index = index_list.tolist().index(index)
            else:
                raise TypeError("index_list must be a list or a numpy ndarray")

            # Generate performance characteristic text
            if isinstance(field_display_var, str):
                if not isinstance(post_process_forces[field_display_var][gen_index], typing.Iterable):
                    display_val = post_process_forces[field_display_var][gen_index]
                else:
                    display_val = post_process_forces[field_display_var][gen_index][1]
            elif isinstance(field_display_var, list):
                display_val = []
                for fdv in field_display_var:
                    if not isinstance(post_process_forces[fdv][gen_index], typing.Iterable):
                        display_val.append(post_process_forces[fdv][gen_index])
                    else:
                        display_val.append(post_process_forces[fdv][gen_index][1])
            else:
                raise TypeError("field_display_var must be either str or list")

            axs.text(x=-0.15, y=0.32, s=field_display_text(index, field_display_var, display_val),
                     fontdict=dict(size=18, family='serif'))

            # Make the color bar
            field_key = "field_0" if "field_0" in quad[1].keys() else "field"
            cbar = fig.colorbar(quad[1][field_key], ax=axs, shrink=0.7)
            cbar.ax.set_ylabel(flow_var_label_matplotlib[var], fontdict=dict(size=18, family='serif'))
            cbar.ax.tick_params(axis='y', which='major', labelsize=14)

            # Set various plot parameters
            axs.set_xlim([-0.2, 1.6])
            axs.set_ylim([-0.4, 0.4])
            format_axis_scientific(axs)
            axs.set_aspect('equal')

            # Label the axes
            axs.set_xlabel(r"$x/c_{main}$", fontdict=dict(size=18, family='serif'))
            axs.set_ylabel(r"$y/c_{main}$", fontdict=dict(size=18, family='serif'))

            # Generate the legend
            proxy_line = Line2D([], [], color=airfoil_color)
            proxy_patch_airfoil = Patch(color="#606060")
            proxy_patch_bl = Patch(color="white")
            proxy_arrow = Line2D([], [], color="blue")
            if self.underwing:
                legend_loc = "upper right"
            else:
                legend_loc = "lower right"
            axs.legend((proxy_line, proxy_patch_airfoil, proxy_patch_bl, proxy_arrow),
                       ('Surface', 'Solid', 'B.L.', 'Stag. S.L.'),
                       prop={'size': 17, 'family': font['family']}, loc=legend_loc, frameon=False)

            # Plot outputs
            fig.set_tight_layout('tight')
            fig_fname_no_ext = os.path.join(self.image_dir, f'field_{index}_{var}{weight_str}')
            for ext in image_extensions:
                kwargs = {'bbox_inches': 'tight'}
                if ext == '.png':
                    kwargs['dpi'] = 600
                fig.savefig(fig_fname_no_ext + ext, **kwargs)

    def generate_gif_from_images(self, var: str, index: int or typing.Iterable = None,
                                 save_name: str = 'opt_history.gif', duration: int = 150, weight_idx: int = 0):
        """Borrowed from https://pythonprogramming.altervista.org/png-to-gif/"""
        index = self.set_index(index)
        weight_str = self.get_weight_str(weight_idx)
        frames = []
        for i in index:
            image = os.path.join(self.image_dir, f'field_{i}_{var}{weight_str}.png')
            new_frame = PIL.Image.open(image)
            frames.append(new_frame)

        # Save into a GIF file that loops forever
        frames[0].save(os.path.join(self.image_dir, save_name), format='GIF',
                       append_images=frames[1:],
                       save_all=True,
                       duration=duration, loop=0)

    def generate_field_gif_matplotlib(self, var: str,
                                      index: int or typing.Iterable = None, shading: str = 'gouraud'):

        flow_var_label_matplotlib = {'M': r'Mach Number',
                                     'Cp': r'Pressure Coefficient',
                                     'p': r'Static Pressure ($p/p_\infty$)',
                                     'rho': r'Density ($\rho/\rho_\infty$)',
                                     'u': r'Velocity-x ($u/V_\infty$)',
                                     'v': r'Velocity-y ($v/V_\infty$)',
                                     'q': r'Speed of Sound ($q/V_\infty$)',
                                     "Cpt": "Total Pressure Coefficient"}

        airfoil_color = 'black'

        post_process_forces = load_data(self.post_process_force_file)

        index = self.set_index(index)

        def animate(i):
            for h in self.coord_plot_handles:
                h_ = h.pop(0)
                h_.remove()
            for h in self.text_handles:
                h.remove()
            for h in self.solid_handles:
                h.remove()
            quad = self.generate_field_matplotlib(var=var, axs=None, index=i, cmap_field="Spectral_r",
                                                  cmap_airfoil=mpl_colors.ListedColormap("gray"), shading=shading,
                                                  vmin=0.0, vmax=1.4)
            coords = load_data(os.path.join(self.analysis_dir, 'coords', f'coords_{i}.json'))
            self.coord_plot_handles = []
            self.solid_handles = []
            self.text_handles = []
            for airfoil in coords:
                airfoil = np.array(airfoil)
                self.coord_plot_handles.append(axs.plot(airfoil[:, 0], airfoil[:, 1], color=airfoil_color))
                polygon = Polygon(airfoil, closed=False, color="#000000AA")
                self.solid_handles.append(axs.add_patch(polygon))
            self.text_handles.append(
                axs.text(x=-0.15, y=0.32, s=fr"Gen. {i}: $C_d = {post_process_forces['Cd'][i][1] * 10000:.1f}$ counts",
                         fontdict=font))
            # TODO: make sure that the color bar updates with each frame if vmin and vmax are set to None ('auto')
            if self.cbar is None:
                # divider = make_axes_locatable(axs)
                # cax = divider.append_axes("right", size="5%", pad=0.15)
                self.cbar = fig.colorbar(quad['field'], ax=axs, shrink=0.6)
                self.cbar.ax.set_ylabel(flow_var_label_matplotlib[var], fontdict=font)

        fig, axs = plt.subplots(figsize=(11, 3.3))
        plt.subplots_adjust(left=0.08, bottom=-0.1, right=1.05, top=1.2, wspace=0.0, hspace=0.0)
        axs.set_xlim([-0.2, 1.8])
        axs.set_ylim([-0.2, 0.4])
        axs.set_aspect('equal')
        axs.set_xlabel(r"$x/c_{main}$", fontdict=font)
        axs.set_ylabel(r"$y/c_{main}$", fontdict=font)
        proxy_line = Line2D([], [], color=airfoil_color)
        proxy_patch_airfoil = Patch(color="#000000DD")
        proxy_patch_bl = Patch(color="#606060")
        axs.legend((proxy_line, proxy_patch_airfoil, proxy_patch_bl), ('Surface', 'Solid', 'BL/Stag.'))

        anim = animation.FuncAnimation(fig, animate, frames=index, blit=False)

        # anim.save(os.path.join(self.image_dir, f'quad_animation_{var}_{shading}_shading.mp4'),
        #           writer=animation.FFMpegWriter(fps=10), dpi=450)

        anim.save(os.path.join(self.image_dir, f'quad_animation_{var}_{shading}_shading.gif'),
                  writer=animation.PillowWriter(fps=10))



from copy import deepcopy
from abc import ABC, abstractmethod


class DualRep:

    # These attributes are non-serializable because they contain references to PyQt5 graphics or signal/slot objects,
    # which are inherently non-serializable
    non_serializable_attributes = ["tree", "tree_item", "canvas", "canvas_item", "gui_obj"]

    def __deepcopy__(self, memo):
        """
        Overwrite the ``deepcopy`` method to set any non-serializable attributes in the copied object to ``None``.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():

            # Eliminate any GUI references so the object is deep-copyable
            if k in self.non_serializable_attributes:
                v = None

            if k == "equation_dict" and isinstance(v, dict) and "__builtins__" in v:
                v["__builtins__"] = None

            setattr(result, k, deepcopy(v, memo))
        return result


class PymeadObj(ABC, DualRep):
    """
    Base class for all objects in pymead.
    """

    def __init__(self, sub_container: str):
        """
        Parameters
        ----------
        sub_container: str
            Sub-container where this object will be stored in the ``GeometryCollection``
        """
        self.sub_container = sub_container
        self._name = None
        self.geo_col = None
        self.gcs = None
        self.tree_item = None
        self.canvas_item = None
        self.gui_obj = None
        self.tree = None
        self.canvas = None

    def name(self):
        """
        Retrieves the parameter name

        Returns
        =======
        str
            The parameter name
        """
        return self._name

    def set_name(self, name: str):
        """
        Sets the object name.

        Parameters
        ==========
        name: str
            The object name
        """
        if self.geo_col is not None:
            # Rename the reference in the geometry collection if necessary
            sub_container = self.geo_col.container()[self.sub_container]

            if self.name() in sub_container and sub_container[self.name()] is self:
                sub_container[name] = sub_container[self.name()]
                sub_container.pop(self.name())

        self._name = name

    @abstractmethod
    def get_dict_rep(self) -> dict:
        """
        Gets a dictionary representation of the pymead object. In general, this dictionary should consist of only
        the required arguments for object instantiation. For example, the dictionary representation of a point looks
        something like this: ``{"x": 0.3, "y": 0.5}``. If the argument requires a reference to a ``PymeadObj``
        rather than a string or float value, the ``name()`` method should be the value that is stored. For
        an example, see the overridden value of this method in ``pymead.core.airfoil.Airfoil``.
        All subclasses of ``PymeadObj`` must implement this method, since it is the way pymead objects are stored in
        saved instances of a ``GeometryCollection`` (``.jmea`` files).
        """
        pass



import typing

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QCloseEvent
from PyQt5.QtWidgets import QMainWindow, QDockWidget, QGridLayout, QApplication, QWidget
from PyQt5.QtCore import pyqtSignal


class PymeadDockWidget(QDockWidget):
    """
    Simple re-implementation of QDockWidget with special handling of tab close with a custom signal.
    """
    tab_closed = pyqtSignal(str, QCloseEvent)

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Sends a signal that the tab has been closed along with the name of the tab and the QCloseEvent object.
        """
        self.tab_closed.emit(f"{self.windowTitle()}", event)


class DockableTabWidget(QMainWindow):
    tab_closed = pyqtSignal(str, QCloseEvent)

    def __init__(self, parent=None, cancel_if_tab_name_exists: bool = False):
        super().__init__(parent=parent)
        self.setDockNestingEnabled(True)
        self.w = QWidget()
        layout = QGridLayout()
        self.dock_widgets = []
        self.names = []
        self.w.setLayout(layout)
        self.setCentralWidget(self.w)
        self.first_dock_widget = None
        self.setWindowTitle('Geometry & Analysis')
        self.cancel_if_tab_name_exists = cancel_if_tab_name_exists
        self.tabifiedDockWidgetActivated.connect(self.activated)

    def add_new_tab_widget(self, widget, name, connect_func: typing.Callable):
        if not (self.cancel_if_tab_name_exists and name in self.names):
            dw = PymeadDockWidget(name, self)
            dw.setWidget(widget)
            dw.setFloating(False)
            dw.tab_closed.connect(self.on_tab_closed)
            self.dock_widgets.append(dw)
            self.names.append(name)
            if len(self.dock_widgets) == 2:
                self.addDockWidget(Qt.LeftDockWidgetArea, self.dock_widgets[-2])
                self.addDockWidget(Qt.RightDockWidgetArea, dw)
                self.setCentralWidget(QWidget())
                self.tabifyDockWidget(self.dock_widgets[-2], self.dock_widgets[-1])
                self.splitDockWidget(self.dock_widgets[-2], self.dock_widgets[-1], Qt.Horizontal)
            elif len(self.dock_widgets) == 3:
                self.addDockWidget(Qt.BottomDockWidgetArea, dw)
                self.tabifyDockWidget(self.dock_widgets[-2], self.dock_widgets[-1])
                self.splitDockWidget(self.dock_widgets[-2], self.dock_widgets[-1], Qt.Vertical)
            elif len(self.dock_widgets) > 3:
                self.addDockWidget(Qt.RightDockWidgetArea, dw)
                self.tabifyDockWidget(self.dock_widgets[-2], self.dock_widgets[-1])
            else:
                self.setCentralWidget(dw)
            self.tab_closed.connect(connect_func)

    def on_tab_closed(self, name: str, event: QCloseEvent):
        if name == "Geometry":
            event.ignore()
            return
        idx = self.names.index(name)
        self.names.pop(idx)
        self.dock_widgets.pop(idx)
        self.tab_closed.emit(name, event)

    def activated(self, dw: QDockWidget):
        self.current_dock_widget = dw
